import datetime
import os
import shutil
import time
import unittest
import warnings

import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.decomposition import PCA
from sklearn.dummy import DummyRegressor, DummyClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import KFold
from sklearn.pipeline import Pipeline as SKLPipeline
from sklearn.preprocessing import StandardScaler
from sklearn.inspection import permutation_importance
from keras.metrics import Accuracy

from photonai.base import PipelineElement, Hyperpipe, OutputSettings, Preprocessing, CallbackElement, Branch, Stack, \
    Switch, ParallelBranch
from photonai.optimization import IntegerRange, Categorical
from photonai.optimization.optimization_info import Optimization
from photonai.processing.results_handler import ResultsHandler
from photonai.processing.results_structure import MDBConfig, MDBFoldMetric, \
    MDBInnerFold, MDBOuterFold, MDBScoreInformation, MDBDummyResults, MDBHyperpipe
from photonai.helper.dummy_elements import DummyTransformer, DummyYAndCovariatesTransformer
from photonai.helper.photon_base_test import elements_to_dict, PhotonBaseTest


class HyperpipeTests(PhotonBaseTest):

    def setup_hyperpipe(self, output_settings=None):
        self.hyperpipe = Hyperpipe('god', inner_cv=self.inner_cv_object,
                                   metrics=self.metrics,
                                   best_config_metric=self.best_config_metric,
                                   project_folder=self.tmp_folder_path,
                                   output_settings=output_settings,
                                   verbosity=0)
        self.hyperpipe += self.ss_pipe_element
        self.hyperpipe += self.pca_pipe_element
        self.hyperpipe.add(self.svc_pipe_element)

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(HyperpipeTests, cls).setUpClass()

    def setUp(self):

        super(HyperpipeTests, self).setUp()
        self.ss_pipe_element = PipelineElement('StandardScaler')
        self.pca_pipe_element = PipelineElement('PCA', {'n_components': [1, 2]}, random_state=42, test_disabled=True)
        self.svc_pipe_element = PipelineElement('SVC', {'C': [0.1, 1], 'kernel': ['linear']}, # 'rbf', 'sigmoid']
                                                random_state=42)

        self.inner_cv_object = KFold(n_splits=3)
        self.metrics = ["accuracy", 'recall', 'precision']
        self.best_config_metric = "accuracy"
        self.setup_hyperpipe()

        dataset = load_breast_cancer()
        self.__X = dataset.data
        self.__y = dataset.target

    def test_init(self):
        # test that all init parameters can be retrieved via the cleaned up subclasses
        self.assertEqual(self.hyperpipe.name, 'god')

        # in case don't give information, check for the default parameters, otherwise for the infos given in setUp
        # Cross Validation
        self.assertIsNotNone(self.hyperpipe.cross_validation)
        self.assertEqual(self.hyperpipe.cross_validation.inner_cv, self.inner_cv_object)
        self.assertIsNone(self.hyperpipe.cross_validation.outer_cv, None)
        self.assertTrue(self.hyperpipe.cross_validation.use_test_set)
        self.assertTrue(self.hyperpipe.cross_validation.calculate_metrics_per_fold)
        self.assertFalse(self.hyperpipe.cross_validation.calculate_metrics_across_folds)
        self.assertIsNone(self.hyperpipe.cross_validation.outer_folds)
        self.assertDictEqual(self.hyperpipe.cross_validation.inner_folds, {})

        # Optimization
        self.assertIsNotNone(self.hyperpipe.optimization)
        self.assertListEqual(self.hyperpipe.optimization.metrics, self.metrics)
        self.assertEqual(self.hyperpipe.optimization.best_config_metric, self.best_config_metric)
        self.assertEqual(self.hyperpipe.optimization.optimizer_input_str, "grid_search")
        self.assertTrue(self.hyperpipe.optimization.maximize_metric)
        self.assertIsNone(self.hyperpipe.optimization.performance_constraints)
        self.assertDictEqual(self.hyperpipe.optimization.optimizer_params, {})

    def test_add(self):
        # assure pipeline has two elements, first the pca and second the svc
        self.assertEqual(len(self.hyperpipe.elements), 3)
        self.assertIs(self.hyperpipe.elements[0], self.ss_pipe_element)
        self.assertIs(self.hyperpipe.elements[1], self.pca_pipe_element)
        self.assertIs(self.hyperpipe.elements[2], self.svc_pipe_element)
        # todo : assure that no two elements can be added with the same name

        # test add method special cases
        with self.assertRaises(TypeError):
            self.hyperpipe.add(object())

        # assure that preprocessing is identified and set to the extra variable, there is only one preprocessing item
        my_preproc = Preprocessing()
        self.hyperpipe.add(my_preproc)
        self.assertEqual(my_preproc, self.hyperpipe.preprocessing)
        # make sure the element does not end up in the main pipeline
        self.assertTrue([item is not my_preproc for item in self.hyperpipe.elements])

        def my_func(X, y, **kwargs):
            return True
        # test adding callback item
        my_call_back_item = CallbackElement('test_element', my_func, 'predict')
        self.hyperpipe.add(my_call_back_item)
        self.assertIs(self.hyperpipe.elements[-1], my_call_back_item)

    def test_sanity(self):
        # make sure that no metrics means raising an error
        with self.assertRaises(ValueError):
            Hyperpipe("hp_name", inner_cv=self.inner_cv_object)

        # make sure that if no best config metric is given, PHOTON raises a warning
        # with self.assertWarns(Warning) as w:
        # with warnings.catch_warnings(record=True) as w:
        with self.assertRaises(Warning) as w:
            Hyperpipe("hp_name", inner_cv=self.inner_cv_object, metrics=["accuracy", "f1_score"])
            assert any("No best config metric was given" in s for s in [e.message.args[0] for e in w])

        # with warnings.catch_warnings(record=True) as w:
        with self.assertRaises(Warning) as w:
            Hyperpipe("hp_name", inner_cv=self.inner_cv_object, best_config_metric=["accuracy", "f1_score"])
            assert any("Best Config Metric must be a single" in s for s in [e.message.args[0] for e in w])

        with self.assertRaises(NotImplementedError):
            Hyperpipe("hp_name", inner_cv=self.inner_cv_object,
                                 best_config_metric='accuracy', metrics=["accuracy"],
                                 calculate_metrics_across_folds=False,
                                 calculate_metrics_per_fold=False)

        with self.assertRaises(AttributeError):
            Hyperpipe("hp_name", best_config_metric='accuracy', metrics=["accuracy"])

        data = np.random.random((500, 50))

        with self.assertRaises(ValueError):
            targets = np.random.randint(0, 1, (500, 2))
            self.hyperpipe.fit(data, targets)

    def test_hyperpipe_with_custom_metric(self):

        def custom_metric(y_true, y_pred):
            return 99.9

        self.hyperpipe = Hyperpipe('god', inner_cv=self.inner_cv_object,
                                   metrics=[('custom_metric', custom_metric), 'accuracy'],
                                   best_config_metric=Accuracy,
                                   project_folder=self.tmp_folder_path)
        self.hyperpipe += self.ss_pipe_element
        self.hyperpipe.add(self.svc_pipe_element)
        self.hyperpipe.fit(self.__X, self.__y)

        self.assertTrue('custom_metric' in self.hyperpipe.results.best_config.best_config_score.validation.metrics)
        self.assertEqual(self.hyperpipe.results.best_config.best_config_score.validation.metrics['custom_metric'], 99.9)

        expected_num_of_metrics = len(self.hyperpipe.optimization.metrics)
        # one: accuracy, two: custom metric registered as "custom_metric", three: keras Metric registered as function
        self.assertEqual(expected_num_of_metrics, 3)

        # dummy average values
        self.assertTrue(len(self.hyperpipe.results.dummy_estimator.metrics_train), expected_num_of_metrics)
        self.assertTrue(len(self.hyperpipe.results.dummy_estimator.metrics_test), expected_num_of_metrics)

        # overall average values
        self.assertTrue(len(self.hyperpipe.results.metrics_train), 2 * expected_num_of_metrics)
        self.assertTrue(len(self.hyperpipe.results.metrics_test), 2 * expected_num_of_metrics)

    def test_preprocessing(self):

        prepro_pipe = Preprocessing()
        prepro_pipe += PipelineElement.create("dummy", DummyYAndCovariatesTransformer(), {})

        self.hyperpipe += prepro_pipe
        self.hyperpipe.fit(self.__X, self.__y)

        self.assertTrue(np.array_equal(self.__y + 1, self.hyperpipe.data.y))

    def test_permutation_feature_importances(self):
        hp = Hyperpipe('god',
                       outer_cv=KFold(n_splits=3),
                       inner_cv=KFold(n_splits=2),
                       metrics=self.metrics,
                       best_config_metric=self.best_config_metric,
                       project_folder=self.tmp_folder_path,
                       verbosity=0)
        svc = PipelineElement('SVC')
        hp += svc
        hp.fit(self.__X, self.__y)

        score_photon = hp.score(self.__X, self.__y)
        score_element = svc.score(self.__X, self.__y)
        self.assertAlmostEqual(score_photon, score_element)

        # do it on outer folds
        permutation_list_outer = hp._calculate_permutation_importances(n_repeats=5, random_state=0)
        self.assertEqual(len(permutation_list_outer["mean"]), 3)

        permutation_score_outer = hp.get_permutation_feature_importances(n_repeats=5, random_state=0)
        self.assertTrue("mean" in permutation_score_outer)
        self.assertTrue("std" in permutation_score_outer)
        self.assertEqual(permutation_score_outer["mean"].shape, (self.__X.shape[1],))
        self.assertEqual(permutation_score_outer["std"].shape, (self.__X.shape[1],))

        # do it on inner folds but on training sets from outer split
        hp.cross_validation.use_test_set = False
        hp.fit(self.__X, self.__y)
        permutation_list_inner = hp._calculate_permutation_importances(n_repeats=5)
        self.assertEqual(len(permutation_list_inner["mean"]), 3*2)
        permutation_score_inner = hp.get_permutation_feature_importances(n_repeats=5)
        self.assertEqual(permutation_score_inner["mean"].shape, (self.__X.shape[1],))
        self.assertEqual(permutation_score_inner["std"].shape, (self.__X.shape[1],))
        # check that validation set permutation importances (inner folds) differ from those of test set (outer folds)
        self.assertFalse(np.array_equal(permutation_score_outer["mean"], permutation_score_inner["mean"]))

        # do it on inner folds only
        hp.cross_validation.outer_folds = {}
        hp.cross_validation.outer_cv = None
        hp.cross_validation.use_test_set = False
        hp.fit(self.__X, self.__y)
        permutation_list_no_outer = hp._calculate_permutation_importances(n_repeats=5)
        self.assertEqual(len(permutation_list_no_outer), 2)
        permutation_score_no_outer = hp.get_permutation_feature_importances(n_repeats=5)
        self.assertEqual(permutation_score_inner["mean"].shape, (self.__X.shape[1],))
        self.assertEqual(permutation_score_inner["std"].shape, (self.__X.shape[1],))

        # raise error
        def fake_metric(y_true, y_pred):
           return 'a'

        hp = Hyperpipe('god',
                       outer_cv=KFold(n_splits=3),
                       inner_cv=KFold(n_splits=2),
                       metrics=[('fake_metric', fake_metric)],
                       best_config_metric='fake_metric',
                       project_folder=self.tmp_folder_path,
                       verbosity=0)
        svc = PipelineElement('SVC')
        hp += svc
        try:
            hp.fit(self.__X, self.__y)
        except Exception as e:
            # should produce an error so that hp.results.best_config is None.
            pass
        with self.assertRaises(ValueError):
            hp.get_permutation_feature_importances(n_repeats=5)
        with self.assertRaises(ValueError):
            hp._calculate_permutation_importances(n_repeats=5)

    def test_estimation_type(self):
        def callback(X, y=None, **kwargs):
            pass

        pipe = Hyperpipe('name', inner_cv=KFold(n_splits=2), best_config_metric='mean_squared_error')

        with self.assertRaises(NotImplementedError):
            pipe += PipelineElement('PCA')
            est_type = pipe.estimation_type

        pipe += PipelineElement('SVC')
        self.assertEqual(pipe.estimation_type, 'classifier')

        pipe.elements[-1] = PipelineElement('SVR')
        self.assertEqual(pipe.estimation_type, 'regressor')

        with self.assertRaises(NotImplementedError):
            pipe.elements[-1] = CallbackElement('MyCallback', callback)
            est_type = pipe.estimation_type

    def test_copy_me(self):
        self.maxDiff = None
        copy = self.hyperpipe.copy_me()
        copy2 = self.hyperpipe.copy_me()
        self.assertDictEqual(elements_to_dict(copy), elements_to_dict(self.hyperpipe))

        copy_after_fit = self.hyperpipe.fit(self.__X, self.__y).copy_me()

        copy_after_fit = elements_to_dict(copy_after_fit)
        # the current_configs of the elements are not None after calling fit() on a hyperpipe
        # when copying the respective PipelineElement, these current_configs are copied, too
        # this is why we need to delete _pipe and elements before asserting for equality
        copy_after_fit['_pipe'] = None
        copy_after_fit['elements'] = None
        copy = elements_to_dict(copy)
        copy['_pipe'] = None
        copy['elements'] = None
        self.assertDictEqual(copy, copy_after_fit)

        # check if deepcopy worked
        copy2.cross_validation.inner_cv.n_splits = 10
        self.assertEqual(copy2.cross_validation.inner_cv.n_splits, 10)
        self.assertEqual(self.hyperpipe.cross_validation.inner_cv.n_splits, 3)

    def test_save_optimum_pipe(self):
        tmp_path = os.path.join(self.tmp_folder_path, 'optimum_pipypipe')
        settings = OutputSettings(overwrite_results=True)

        my_pipe = Hyperpipe('hyperpipe',
                            optimizer='random_grid_search',
                            optimizer_params={'n_configurations': 3},
                            metrics=['accuracy', 'precision', 'recall'],
                            best_config_metric='f1_score',
                            outer_cv=KFold(n_splits=2),
                            inner_cv=KFold(n_splits=2),
                            verbosity=0,
                            project_folder=tmp_path,
                            output_settings=settings)

        preproc = Preprocessing()
        preproc += PipelineElement('StandardScaler')

        # BRANCH WITH QUANTILTRANSFORMER AND DECISIONTREECLASSIFIER
        tree_qua_branch = Branch('tree_branch')
        tree_qua_branch += PipelineElement('QuantileTransformer')
        tree_qua_branch += PipelineElement('DecisionTreeClassifier', {'min_samples_split': IntegerRange(2, 4)},
                                           criterion='gini')

        # BRANCH WITH MinMaxScaler AND DecisionTreeClassifier
        svm_mima_branch = Branch('svm_branch')
        svm_mima_branch += PipelineElement('MinMaxScaler')
        svm_mima_branch += PipelineElement('SVC', {'kernel': Categorical(['rbf', 'linear']), 'C': 2.0}, gamma='auto')

        # BRANCH WITH StandardScaler AND KNeighborsClassifier
        knn_sta_branch = Branch('neighbour_branch')
        knn_sta_branch += PipelineElement.create("dummy", DummyTransformer(), {})
        knn_sta_branch += PipelineElement('KNeighborsClassifier')

        my_pipe += preproc
        # voting = True to mean the result of every branch
        my_pipe += Stack('final_stack', [tree_qua_branch, svm_mima_branch, knn_sta_branch])

        my_pipe += PipelineElement('LogisticRegression', solver='lbfgs')

        my_pipe.fit(self.__X, self.__y)
        model_path = os.path.join(my_pipe.output_settings.results_folder, 'photon_best_model.photon')
        self.assertTrue(os.path.exists(model_path))

        # now move optimum pipe to new folder
        test_folder = os.path.join(my_pipe.output_settings.results_folder, 'new_test_folder')
        new_model_path = os.path.join(test_folder, 'photon_best_model.photon')
        os.makedirs(test_folder)
        shutil.copyfile(model_path, new_model_path)

        # check if load_optimum_pipe also works
        # check if we have the meta information recovered
        loaded_optimum_pipe = Hyperpipe.load_optimum_pipe(new_model_path)
        self.assertIsNotNone(loaded_optimum_pipe._meta_information)
        self.assertIsNotNone(loaded_optimum_pipe._meta_information['photon_version'])

        # check if predictions stay realiably the same
        y_pred_loaded = loaded_optimum_pipe.predict(self.__X)
        y_pred = my_pipe.optimum_pipe.predict(self.__X)
        np.testing.assert_array_equal(y_pred_loaded, y_pred)

    def test_save_optimum_pipe_custom_element(self):
        tmp_path = os.path.join(self.tmp_folder_path, 'optimum_pipypipe')
        settings = OutputSettings(overwrite_results=True)

        my_pipe = Hyperpipe('hyperpipe',
                            optimizer='random_grid_search',
                            optimizer_params={'n_configurations': 1},
                            metrics=['accuracy', 'precision', 'recall'],
                            best_config_metric='f1_score',
                            outer_cv=KFold(n_splits=2),
                            inner_cv=KFold(n_splits=2),
                            verbosity=0,
                            project_folder=tmp_path,
                            output_settings=settings)
        my_pipe += PipelineElement('KerasDnnClassifier', {}, epochs=1, hidden_layer_sizes=[5])
        my_pipe.fit(self.__X, self.__y)
        model_path = os.path.join(my_pipe.output_settings.results_folder, 'photon_best_model.photon')
        self.assertTrue(os.path.exists(model_path))

        # check if load_optimum_pipe also works
        # check if we have the meta information recovered
        loaded_optimum_pipe = Hyperpipe.load_optimum_pipe(model_path)
        self.assertIsNotNone(loaded_optimum_pipe._meta_information)

    def test_failure_to_save_optimum_pipe(self):
        tmp_path = os.path.join(self.tmp_folder_path, 'optimum_pipypipe')
        settings = OutputSettings(overwrite_results=True)

        my_pipe = Hyperpipe('hyperpipe',
                            optimizer='random_grid_search',
                            optimizer_params={'n_configurations': 1},
                            metrics=['accuracy', 'precision', 'recall'],
                            best_config_metric='f1_score',
                            outer_cv=KFold(n_splits=2),
                            inner_cv=KFold(n_splits=2),
                            verbosity=0,
                            project_folder=tmp_path,
                            output_settings=settings)
        my_pipe += PipelineElement('KNeighborsClassifier')
        my_pipe.fit(self.__X, self.__y)
        model_path = os.path.join(my_pipe.output_settings.results_folder, 'photon_best_model_wrong_path.photon')
        with self.assertRaises(FileNotFoundError):
            Hyperpipe.load_optimum_pipe(model_path)

    def test_overwrite_result_folder(self):
        """
        Test for right handling of parameter output_settings.overwrite.
        """
        def get_summary_file():
            return os.path.join(self.hyperpipe.output_settings.results_folder, 'photon_summary.txt')

        # Case 1: default
        output_settings1 = OutputSettings(save_output=True,
                                          overwrite_results=False)
        self.setup_hyperpipe(output_settings1)
        self.hyperpipe.fit(self.__X, self.__y)
        tmp_path = get_summary_file()

        time.sleep(2)

        # again with same settings
        self.setup_hyperpipe(output_settings1)
        self.hyperpipe.fit(self.__X, self.__y)
        tmp_path2 = get_summary_file()

        # we expect a new output folder each time with timestamp
        self.assertNotEqual(tmp_path, tmp_path2)

        # Case 2 overwrite results: all in the same folder
        output_settings2 = OutputSettings(save_output=True,
                                          overwrite_results=True)
        self.setup_hyperpipe(output_settings2)
        self.hyperpipe.fit(self.__X, self.__y)
        tmp_path = get_summary_file()
        tmp_date = os.path.getmtime(tmp_path)

        self.setup_hyperpipe(output_settings2)
        self.hyperpipe.fit(self.__X, self.__y)
        tmp_path2 = get_summary_file()
        tmp_date2 = os.path.getmtime(tmp_path2)

        # same folder but summary file is overwritten through the new analysis
        self.assertEqual(tmp_path, tmp_path2)
        self.assertNotEqual(tmp_date, tmp_date2)

        # Case 3: we have a cache folder
        self.hyperpipe.cache_folder = self.cache_folder_path
        shutil.rmtree(self.cache_folder_path, ignore_errors=True)
        self.hyperpipe.fit(self.__X, self.__y)
        self.assertTrue(os.path.exists(self.cache_folder_path))

    def test_random_state(self):
        self.hyperpipe.random_state = 4567
        self.hyperpipe.fit(self.__X, self.__y)
        # assure we spread the word.. !
        self.assertEqual(self.hyperpipe.random_state, 4567)
        self.assertEqual(self.hyperpipe._pipe.random_state, 4567)
        self.assertEqual(self.hyperpipe.optimum_pipe.random_state, 4567)
        self.assertEqual(self.hyperpipe._pipe.elements[-1][-1].random_state, 4567)
        self.assertEqual(self.hyperpipe._pipe.elements[-1][-1].base_element.random_state, 4567)


    def test_dummy_estimator_preparation(self):

        self.hyperpipe.results = MDBHyperpipe()
        self.hyperpipe.results.dummy_estimator = dummy_estimator = MDBDummyResults()

        # one time regressor, one time classifier, one time strange object
        self.hyperpipe.elements = list()
        self.hyperpipe.add(PipelineElement('SVC'))
        dummy_estimator = self.hyperpipe._prepare_dummy_estimator()
        self.assertTrue(isinstance(dummy_estimator, DummyClassifier))

        self.hyperpipe.elements = list()
        self.hyperpipe.add(PipelineElement('SVR'))
        dummy_estimator = self.hyperpipe._prepare_dummy_estimator()
        self.assertTrue(isinstance(dummy_estimator, DummyRegressor))

        with self.assertRaises(NotImplementedError):
            self.hyperpipe.elements = list()
            self.hyperpipe.add(PipelineElement('PCA'))
            dummy_estimator = self.hyperpipe._prepare_dummy_estimator()
            self.assertIsNone(dummy_estimator)

    def setup_crazy_pipe(self):
        # erase all, we need a complex and crazy task
        self.hyperpipe.elements = list()

        nmb_list = list()
        for i in range(5):
            nmb = ParallelBranch(name=str(i), nr_of_processes=i+3)
            sp = PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(1, 50)})
            nmb += sp
            nmb_list.append(nmb)

        my_switch = Switch('disabling_test_switch')
        my_switch += nmb_list[0]
        my_switch += nmb_list[1]

        my_stack = Stack('stack_of_branches')
        for i in range(3):
            my_branch = Branch('branch_' + str(i+2))
            my_branch += PipelineElement('StandardScaler')
            my_branch += nmb_list[i+2]
            my_stack += my_branch

        self.hyperpipe.add(my_stack)
        self.hyperpipe.add(PipelineElement('StandardScaler'))
        self.hyperpipe.add(my_switch)
        self.hyperpipe.add(PipelineElement('SVC'))
        return nmb_list

    def test_eval_final_performance_deprecation(self):
        with self.assertRaises(DeprecationWarning):
            Hyperpipe('name', eval_final_performance=True)

    def test_recursive_disabling(self):
        list_of_elements_to_detect = self.setup_crazy_pipe()
        self.hyperpipe._pipe = Branch.prepare_photon_pipe(list_of_elements_to_detect)
        Hyperpipe.disable_multiprocessing_recursively(self.hyperpipe._pipe)
        self.assertTrue([i.nr_of_processes == 1 for i in list_of_elements_to_detect])

    def test_recursive_cache_folder_propagation(self):
        list_of_elements = self.setup_crazy_pipe()
        self.hyperpipe._pipe = Branch.prepare_photon_pipe(self.hyperpipe.elements)
        self.hyperpipe.recursive_cache_folder_propagation(self.hyperpipe._pipe, self.cache_folder_path, 'fold_id_123')
        for i, nmbranch in enumerate(list_of_elements):
            if i > 1:
                start_folder = os.path.join(self.cache_folder_path, 'branch_' + nmbranch.name)
            else:
                start_folder = self.cache_folder_path
            expected_folder = os.path.join(start_folder, nmbranch.name)
            self.assertEqual(nmbranch.base_element.cache_folder, expected_folder)

    def test_setup_error_file(self):
        # when we init the hyperpipe the file should exist
        self.assertTrue(os.path.isfile(self.hyperpipe.output_settings.setup_error_file))
        # when we call fit it shall disappear
        # we call it empty so that the computation does not occur
        with self.assertRaises(ValueError):
            self.hyperpipe.fit(None, None)
        # however the file should be gone by now
        self.assertFalse(os.path.isfile(self.hyperpipe.output_settings.setup_error_file))


    def test_prepare_result_logging(self):
        # test that results object is given and entails hyperpipe infos
        rfc = PipelineElement('RandomForestClassifier')
        lsvc = PipelineElement('LinearSVC')
        branch = Branch('dummy_branch')
        branch += PipelineElement('SVC')
        self.hyperpipe += Stack('final_stack', [PipelineElement('SVC'), rfc, branch])
        self.hyperpipe += lsvc

        self.hyperpipe.data.X = self.__X
        self.hyperpipe.data.y = self.__y
        self.hyperpipe._prepare_result_logging(datetime.datetime.now())
        self.assertTrue(isinstance(self.hyperpipe.results, MDBHyperpipe))
        self.assertTrue(isinstance(self.hyperpipe.results_handler, ResultsHandler))
        self.assertTrue(len(self.hyperpipe.results.outer_folds) == 0)

        expected_pipeline_struct = {'StandardScaler': str(type(self.ss_pipe_element.base_element)),
                                    'PCA': str(type(self.pca_pipe_element.base_element)),
                                    'SVC': str(type(self.svc_pipe_element.base_element)),
                                    'STACK:final_stack': {'SVC': str(type(self.svc_pipe_element.base_element)),
                                                          'RandomForestClassifier': str(type(rfc.base_element)),
                                                          'BRANCH:dummy_branch': {
                                                              'SVC': str(type(self.svc_pipe_element.base_element))
                                                          }},
                                    'LinearSVC': str(type(lsvc.base_element))
                                    }
        self.assertDictEqual(self.hyperpipe.results.hyperpipe_info.elements, expected_pipeline_struct)

    def test_finalize_optimization(self):
        # it is kind of difficult to test that's why we fake it
        self.hyperpipe.fit(self.__X, self.__y)

        # reset all infos
        self.hyperpipe.results.dummy_estimator.metrics_train = MDBScoreInformation()
        self.hyperpipe.results.dummy_estimator.metrics_test = MDBScoreInformation()
        self.hyperpipe.results.metrics_train = {}
        self.hyperpipe.results.metrics_test = {}
        self.hyperpipe.best_config = None
        self.hyperpipe.results.best_config = MDBConfig()
        self.hyperpipe.optimum_pipe = None

        # now generate infos again
        self.hyperpipe._finalize_optimization()

        expected_num_of_metrics = len(self.hyperpipe.optimization.metrics)
        # dummy average values
        self.assertTrue(len(self.hyperpipe.results.dummy_estimator.metrics_train), expected_num_of_metrics)
        self.assertTrue(len(self.hyperpipe.results.dummy_estimator.metrics_test), expected_num_of_metrics)
        # overall average values
        self.assertTrue(len(self.hyperpipe.results.metrics_train), 2 * expected_num_of_metrics)
        self.assertTrue(len(self.hyperpipe.results.metrics_test), 2 * expected_num_of_metrics)
        # find best config
        self.assertIsNotNone(self.hyperpipe.best_config)
        self.assertIsNotNone(self.hyperpipe.results.best_config)
        self.assertEqual(self.hyperpipe.best_config, self.hyperpipe.results.best_config.config_dict)
        # set optimum pipe and params, # todo: test add preprocessing
        self.assertIsNotNone(self.hyperpipe.optimum_pipe)
        self.assertEqual(self.hyperpipe.optimum_pipe.named_steps["SVC"].base_element.C, self.hyperpipe.best_config["SVC__C"])
        # save optimum model
        self.assertTrue(os.path.isfile(os.path.join(self.hyperpipe.output_settings.results_folder,
                                                    'photon_best_model.photon')))

        # backmapping
        # because the pca is test disabled, we expect the number of features
        self.assertEqual(len(self.hyperpipe.results.best_config_feature_importances[0]), self.__X.shape[1])
        backmapped_feature_importances = os.path.join(self.hyperpipe.output_settings.results_folder,
                                                      'optimum_pipe_feature_importances_backmapped.csv')
        self.assertTrue(os.path.isfile(backmapped_feature_importances))
        loaded_array = np.loadtxt(open(backmapped_feature_importances, 'rb'), delimiter=",")
        self.assertEqual(loaded_array.shape[0], self.__X.shape[1])

    def test_finalize_optimization_preprocessing(self):
        self.hyperpipe.elements = list()

        pre_proc = Preprocessing()
        pre_proc += PipelineElement('StandardScaler')
        self.hyperpipe.add(pre_proc)
        self.hyperpipe.add(PipelineElement('SVC'))
        self.hyperpipe.fit(self.__X, self.__y)

        self.assertTrue(os.path.isfile(os.path.join(self.hyperpipe.output_settings.results_folder,
                                                    'photon_best_model.photon')))

    def test_finalize_optimization_preprocessing_with_client(self):
        self.hyperpipe.elements = list()

        pb = ParallelBranch(name="ParallelBranch", nr_of_processes=2)
        pb += PipelineElement('LabelEncoder')
        pre_proc = Preprocessing()
        pre_proc += pb
        self.hyperpipe.add(pre_proc)
        self.hyperpipe.add(PipelineElement('SVC'))
        self.hyperpipe.fit(self.__X, self.__y)

        self.assertTrue(os.path.isfile(os.path.join(self.hyperpipe.output_settings.results_folder,
                                                    'photon_best_model.photon')))

    def test_optimum_pipe_predict_and_predict_proba_and_transform(self):
        # find best config and test against sklearn
        self.hyperpipe.elements[-1] = PipelineElement('RandomForestClassifier', {'n_estimators': IntegerRange(4, 20,
                                                                                                              step=2)},
                                                      random_state=42)
        self.hyperpipe.fit(self.__X, self.__y)

        # the best config is without PCA so we test it

        best_config_copy = dict(self.hyperpipe.best_config)
        del best_config_copy["PCA__disabled"]
        if self.hyperpipe.best_config["PCA__disabled"]:
            sk_elements = [('StandardScaler', StandardScaler()),
                           ('RandomForestClassifier', RandomForestClassifier(random_state=42))]
        else:
            sk_elements = [('StandardScaler', StandardScaler()),
                           ('PCA', PCA(random_state=42)),
                           ('RandomForestClassifier', RandomForestClassifier(random_state=42))]
        self.sklearn_pipe = SKLPipeline(sk_elements)
        self.sklearn_pipe.set_params(**best_config_copy)
        self.sklearn_pipe.fit(self.__X, self.__y)

        self.assertTrue(np.array_equal(self.sklearn_pipe.predict(self.__X), self.hyperpipe.predict(self.__X)))
        self.assertTrue(np.array_equal(self.sklearn_pipe.predict_proba(self.__X),
                                       self.hyperpipe.predict_proba(self.__X)))
        # fake transform on sklearn pipe
        step1 = self.sklearn_pipe.named_steps["StandardScaler"].transform(self.__X)
        if "PCA" in self.sklearn_pipe.named_steps:
            step2 = self.sklearn_pipe.named_steps["PCA"].transform(self.__X)
        else:
            step2 = step1
        self.assertTrue(np.allclose(step2, self.hyperpipe.transform(self.__X)))


class HyperpipeOptimizationClassTests(unittest.TestCase):

    def test_best_config_metric(self):
        my_pipe_optimizer = Optimization('grid_search', {}, [], 'balanced_accuracy', None)
        self.assertTrue(my_pipe_optimizer.maximize_metric)
        my_pipe_optimizer = Optimization('grid_search', {}, [], 'mean_squared_error', None)
        self.assertFalse(my_pipe_optimizer.maximize_metric)

    def test_optmizer_input_str(self):
        with self.assertRaises(ValueError):
            my_pipe_optimizer = Optimization('unknown_optimizer', {}, [], 'accuracy', None)

        for name, opt_class in Optimization.OPTIMIZER_DICTIONARY.items():
            def get_optimizer(name, params={}):
                my_pipe_optimizer = Optimization(name, params, [], 'accuracy', None)
                return my_pipe_optimizer.get_optimizer()

            if name == 'smac':
                try:
                    import smac
                except ModuleNotFoundError:
                    with self.assertRaises(ModuleNotFoundError):
                        get_optimizer(name)
            if name =='switch':
                get_optimizer(name, {'name': 'random_grid_search'})
            else:
                self.assertIsInstance(get_optimizer(name), opt_class)

    def test_get_optimum_config(self):
        my_pipe_optimizer = Optimization('grid_search', {}, [], 'balanced_accuracy', None)
        list_of_tested_configs = list()
        metric_default = MDBFoldMetric(metric_name='balanced_accuracy',
                                       operation="mean",
                                       value=0.5)
        metric_best = MDBFoldMetric(metric_name='balanced_accuracy',
                                    operation="mean",
                                    value=0.99)
        metric_filter_best = MDBFoldMetric(metric_name='another_accuracy',
                                           operation="mean",
                                           value=0.75)
        metric_filter_default = MDBFoldMetric(metric_name='another_accuracy',
                                              operation="mean",
                                              value=0.55)
        # we add looser configs, one good config, and one good config that failed
        # and check if the good non-failing config is chosen
        for i in range(10):
            config = MDBConfig()
            # number 5 is the winner
            if i == 5 or i == 8:
                config.metrics_test = [metric_best, metric_filter_default]
                config.config_dict = {'any_key': 'pressed'}
            elif i == 7:
                config.metrics_test = [metric_default, metric_filter_best]
                config.config_dict = {'any_key': 'pressed'}
            else:
                config.metrics_test = [metric_default, metric_filter_default]
                config.config_dict = {'any_key': 'not_pressed'}
            if i == 8:
                config.config_failed = True
            list_of_tested_configs.append(config)

        outer_fold = MDBOuterFold()
        outer_fold.tested_config_list = list_of_tested_configs
        winner_config = outer_fold.get_optimum_config(metric=my_pipe_optimizer.best_config_metric,
                                                      maximize_metric=my_pipe_optimizer.maximize_metric)
        self.assertIs(winner_config, list_of_tested_configs[5])
        self.assertEqual(winner_config.get_test_metric('balanced_accuracy'), 0.99)

        winner_config_filtered = outer_fold.get_optimum_config(metric='another_accuracy', maximize_metric=True,
                                                               dict_filter=('any_key', 'pressed'))
        self.assertEqual(winner_config_filtered.get_test_metric('another_accuracy'), 0.75)

    def test_get_optimum_config_outer_folds(self):
        my_pipe_optimizer = Optimization('grid_search', {}, [], 'balanced_accuracy', None)

        outer_fold_list = list()
        for i in range(10):
            outer_fold = MDBOuterFold()
            outer_fold.best_config = MDBConfig()
            outer_fold.best_config.best_config_score = MDBInnerFold()
            outer_fold.best_config.best_config_score.validation = MDBScoreInformation()
            # again fold 5 wins
            if i == 5:
                outer_fold.best_config.best_config_score.validation.metrics = {'balanced_accuracy': 0.99}
            else:
                outer_fold.best_config.best_config_score.validation.metrics = {'balanced_accuracy': 0.5}
            outer_fold_list.append(outer_fold)

        best_config_outer_folds = my_pipe_optimizer.get_optimum_config_outer_folds(outer_fold_list)
        self.assertEqual(best_config_outer_folds.best_config_score.validation.metrics['balanced_accuracy'], 0.99)
        self.assertIs(best_config_outer_folds, outer_fold_list[5].best_config)



import unittest
import types
import numpy as np
import warnings

from photonai.processing.metrics import Scorer, spearman_correlation, specificity, sensitivity, one_hot_to_binary, \
    pearson_correlation, balanced_accuracy, categorical_accuracy_score, variance_explained_score


class ScorerTest(unittest.TestCase):

    def setUp(self):
        """
        Set up for Scorer Tests.
        """
        self.all_implemented_metrics = Scorer.ELEMENT_DICTIONARY.keys()
        self.some_not_implemented_metrics = ["abc_metric", "photon_metric"]

    def test_create(self):
        """
        Test for method create.
        Should searching for the metric by name and instantiates the according calculation function
        """
        for implemented_metric in self.all_implemented_metrics:
            self.assertIsInstance(Scorer.create(implemented_metric), types.FunctionType)

        for not_implemented_metric in self.some_not_implemented_metrics:
            with self.assertRaises(NameError):
                self.assertIsNone(Scorer.create(not_implemented_metric))

    def test_greater_is_better_distinction(self):
        """
        Test for method greater_is_better_distinction.
        Should return Boolean or raise NotImplementedError.
        """
        for implemented_metric in self.all_implemented_metrics:
            self.assertIn(Scorer.greater_is_better_distinction(implemented_metric), [True, False])

        for not_implemented_metric in self.some_not_implemented_metrics:
            with self.assertRaises(NameError):
                Scorer.greater_is_better_distinction(not_implemented_metric)

    def test_calculate_metrics(self):
        """
        Test for method calculate_metrics.
        Handle all given metrics with a scorer call.
        """
        for implemented_metric in self.all_implemented_metrics:
            self.assertIsInstance(Scorer.calculate_metrics([1, 1, 0, 1],
                                                           [0, 1, 0, 1],
                                                           [implemented_metric])[implemented_metric], float)

        for not_implemented_metric in self.some_not_implemented_metrics:
            with self.assertRaises(NameError):
                np.testing.assert_equal(Scorer.calculate_metrics(
                    [1, 1, 0, 1], [0, 1, 0, 1], [not_implemented_metric])[not_implemented_metric], np.nan)

    def test_doubled_custom_metric(self):

        def custom_metric(y_true, y_pred):
            return 99.9

        Scorer.register_custom_metric(('a_custom_metric', custom_metric))

        with self.assertRaises(ValueError):
            Scorer.register_custom_metric(None)

        with warnings.catch_warnings(record=True) as w:
            Scorer.register_custom_metric(('a_custom_metric', custom_metric))
            assert any("is ambiguous. Please specify metric" in s for s in [e.message.args[0] for e in w])

    def test_keras_metric(self):
        try:
            from keras.metrics import MeanAbsoluteError
            Scorer.register_custom_metric(MeanAbsoluteError)

            Scorer.register_custom_metric('any_weird_metric')
        except ImportError:
            pass

    def test_photonai_metrics(self):
        y_true = np.concatenate((np.ones((200,)), np.zeros((400,))))
        y_pred = np.concatenate((np.ones((100,)), np.zeros((500,))))

        pearson_corr = pearson_correlation(y_true, y_pred)
        self.assertAlmostEqual(pearson_corr, 0.6324555320336789)
        spearman_corr = spearman_correlation(y_true, y_pred)
        self.assertAlmostEqual(spearman_corr, 0.632455532033676)
        cat_acc = categorical_accuracy_score(y_true, y_pred)
        self.assertAlmostEqual(cat_acc, 0.8333333333333334)
        v_explained = variance_explained_score(y_true, y_pred)
        self.assertAlmostEqual(v_explained, 0.4000000000000038)

        sens = sensitivity(y_true, y_pred)
        self.assertEqual(sens, 0.5)
        spec = specificity(y_true, y_pred)
        self.assertEqual(spec, 1.0)
        b_acc = balanced_accuracy(y_true, y_pred)
        self.assertEqual(b_acc, 0.75)

        # get np.nan for multidim
        y_multiclass = np.concatenate((np.ones((200,)), np.zeros((400,)),
                                       np.ones(100,)*3))

        self.assertTrue(np.isnan(sensitivity(y_multiclass, y_multiclass)))
        self.assertTrue(np.isnan(specificity(y_multiclass, y_multiclass)))
        self.assertTrue(np.isnan(balanced_accuracy(y_multiclass, y_multiclass)))

    def test_one_hot_decoding(self):
        y_one_hot = np.stack((np.concatenate((np.ones((100,)), np.zeros((100,)))),
                              np.concatenate((np.zeros((100,)), np.ones((100,))))), axis=1)
        binarized_multidim = one_hot_to_binary(y_one_hot)
        self.assertTrue(np.array_equal(binarized_multidim, np.concatenate((np.zeros((100,)), np.ones((100, ))))))



from sklearn.utils import check_random_state
from sklearn.model_selection._split import _BaseKFold
import numpy as np


class StratifiedKFoldRegression(_BaseKFold):
    """Stratified K-Folds cross-validator for continuous target values (regression tasks).

       Provides train/test indices to split data in train/test sets.

       This cross-validation object is a variation of StratifiedKFold that returns
       stratified folds with regard to a continuous target. The folds are made by first sorting
       samples with respect to y. Subsequently, we step through each consecutive k samples and
       randomly allocate exactly one of them to one of the test folds. For a detailed description
       of the process see: http://scottclowe.com/2016-03-19-stratified-regression-partitions/.

       Read more in the :ref:`User Guide <cross_validation>`.

       Parameters
       ----------
       n_splits : int, default=3
           Number of folds. Must be at least 2.

       shuffle : boolean, optional
           Whether to shuffle each stratification of the data before splitting
           into batches.

       random_state : int, RandomState instance or None, optional, default=None
           If int, random_state is the seed used by the random number generator;
           If RandomState instance, random_state is the random number generator;
           If None, the random number generator is the RandomState instance used
           by `np.random`. Used when ``shuffle`` == True.

       Examples
       --------
       >>> from photonai.processing.cross_validation import StratifiedKFoldRegression
       >>> X = np.array([[1, 2], [3, 4], [5, 6], [1, 2], [3, 4], [5, 6]])
       >>> y = np.array([0, 1, 2, 3, 4, 5])
       >>> skf = StratifiedKFoldRegression(n_splits=2)
       >>> skf.get_n_splits(X, y)
       2
       >>> print(skf)  # doctest: +NORMALIZE_WHITESPACE
       StratifiedKFoldRegression(n_splits=2, random_state=None, shuffle=False)
       >>> for train_index, test_index in skf.split(X, y):
       ...    print("TRAIN:", train_index, "TEST:", test_index)
       ...    X_train, X_test = X[train_index], X[test_index]
       ...    y_train, y_test = y[train_index], y[test_index]
       TRAIN: [1 3 5] TEST: [0 2 4]
       TRAIN: [0 2 4] TEST: [1 3 5]

       Notes
       -----
       All the folds have size ``trunc(n_samples / n_splits)``, the last one has
       the complementary.
       """

    def __init__(self, n_splits=3, shuffle=False, random_state=None):
        super(StratifiedKFoldRegression, self).__init__(n_splits, shuffle, random_state)

    def _make_test_folds(self, X, y=None):
        rng = self.random_state
        n_splits = self.n_splits
        y = np.asarray(y)
        if y.ndim > 1:
            msg = 'Target data has more than one dimension. Must be single vector of continuous values.'
            raise NotImplementedError(msg)
        n_samples = y.shape[0]
        self.n_samples = n_samples
        sort_indices = np.argsort(y)
        min_test_samples_per_fold = int(np.ceil(n_samples / n_splits))

        test_folds = [[] for _ in range(n_splits)]
        current = 0
        for i in range(min_test_samples_per_fold):
            start, stop = current, current + n_splits
            if i+1 < min_test_samples_per_fold:
                subset = sort_indices[start:stop]
            else:
                subset = sort_indices[start:]

            if self.shuffle:
                check_random_state(rng).shuffle(subset)
            for k in range(len(subset)):
                test_folds[k].append(subset[k])
            current = stop
        return test_folds

    def _iter_test_masks(self, X=None, y=None, groups=None):
        test_folds = self._make_test_folds(X, y)
        for i in range(self.n_splits):
            test_mask = np.zeros(self.n_samples, dtype=np.bool)
            test_mask[test_folds[i]] = True
            yield test_mask

    def split(self, X=None, y=None, groups=None):
        return super(StratifiedKFoldRegression, self).split(X, y, groups)



import datetime

import numpy as np
from typing import Union, Generator

from photonai.optimization.base_optimizer import PhotonSlaveOptimizer
from photonai.optimization.config_grid import create_global_config_grid
from photonai.photonlogger.logger import logger


class GridSearchOptimizer(PhotonSlaveOptimizer):
    """Grid search optimizer.

    Searches for the best configuration by iteratively
    testing a grid of possible hyperparameter combinations.

    Example:
        ``` python
        my_pipe = Hyperpipe(name='grid_based_pipe',
                            optimizer='grid_search',
                            ...
                            )
        my_pipe.fit(X, y)
        ```

    """
    def __init__(self):
        """Initialize the object."""
        self.param_grid = []
        self.pipeline_elements = None
        self.parameter_iterable = None
        self.ask = self.next_config_generator()

    def prepare(self, pipeline_elements: list, maximize_metric: bool) -> None:
        """
        Creates a grid from a list of PipelineElements.
        Hyperparameters can be accessed via pipe_element.hyperparameters.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create the hyperparameter space.

            maximize_metric:
                Boolean to distinguish between score and error.

        """
        self.pipeline_elements = pipeline_elements
        self.ask = self.next_config_generator()
        self.param_grid = create_global_config_grid(self.pipeline_elements)
        logger.info("Grid Search generated " + str(len(self.param_grid)) + " configurations")

    def next_config_generator(self) -> Generator:
        """
        Generator for new configs - ask method.

        Returns:
            Yields the next config.

        """
        for parameters in self.param_grid:
            yield parameters


class RandomGridSearchOptimizer(GridSearchOptimizer):
    """Random grid search optimizer.

    Searches for the best configuration by randomly
    testing n points of a grid of possible hyperparameters.

    Example:
        ``` python
        my_pipe = Hyperpipe(name='rgrid_based_pipe',
                            optimizer='random_grid_search',
                            optimizer_params={'n_configurations': 50,
                                              'limit_in_minutes': 10},
                            ...
                            )
        my_pipe.fit(X, y)
        ```

    """
    def __init__(self, limit_in_minutes: Union[float, None] = None, n_configurations: Union[int, None] = 25):
        """
        Initialize the object.

        Parameters:
            limit_in_minutes:
                Total time in minutes.

            n_configurations:
                Number of configurations to be calculated.

        """
        super(RandomGridSearchOptimizer, self).__init__()
        self._k = n_configurations
        self.n_configurations = self._k
        self.limit_in_minutes = limit_in_minutes
        self.start_time, self.end_time = None, None

    def prepare(self, pipeline_elements: list, maximize_metric: bool) -> None:
        """
        Prepare hyperparameter search.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create the hyperparameter space.

            maximize_metric:
                Boolean to distinguish between score and error.

        """
        super(RandomGridSearchOptimizer, self).prepare(pipeline_elements, maximize_metric)
        self.start_time = None
        self.n_configurations = self._k
        self.param_grid = list(self.param_grid)
        # create random order in list
        np.random.shuffle(self.param_grid)
        if self.n_configurations is not None:
            # k is maximal all grid items
            if self.n_configurations > len(self.param_grid):
                self.n_configurations = len(self.param_grid)
            self.param_grid = self.param_grid[0:self.n_configurations]

    def next_config_generator(self) -> Generator:
        """
        Generator for new configs - ask method.

        Returns:
            Yields the next config.

        """
        if self.start_time is None and self.limit_in_minutes is not None:
            self.start_time = datetime.datetime.now()
            self.end_time = self.start_time + datetime.timedelta(minutes=self.limit_in_minutes)
        for parameters in super(RandomGridSearchOptimizer, self).next_config_generator():
            if self.limit_in_minutes is None or datetime.datetime.now() < self.end_time:
                yield parameters



import math
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import StratifiedKFold

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import FloatRange, Categorical


# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = load_breast_cancer(return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe(name='llembedding_pipe',
                    optimizer='grid_search',
                    metrics=['balanced_accuracy'],
                    best_config_metric='balanced_accuracy',
                    outer_cv=StratifiedKFold(n_splits=2, shuffle=True, random_state=42),
                    inner_cv=StratifiedKFold(n_splits=2, shuffle=True, random_state=42),
                    project_folder='./tmp/',
                    verbosity=0)

my_pipe += PipelineElement('StandardScaler')

my_pipe += PipelineElement('LocallyLinearEmbedding', n_components=math.ceil(0.1 * len(X[0])))

my_pipe += PipelineElement('SVC',
                           hyperparameters={'kernel': Categorical(['rbf', 'linear']),
                                            'C': FloatRange(1, 6)},
                           gamma='scale')

my_pipe.fit(X, y)



import json
import sys
import inspect
import numpy as np

from sklearn.model_selection import *
from photonai.optimization.hyperparameters import *
from photonai.base.photon_elements import *

try:
    from photonai.base.hyperpipe import Hyperpipe, OutputSettings
except:
    pass


class JsonTransformer(object):

    def __init__(self, black_list: list = None):
        if black_list is None:
            self.black_list = ["base_element"]
        else:
            self.black_list = black_list
        self.data_json = {}
        self.attribute_allocator = {"PipelineElement": ["initial_name", "initial_hyperparameters",
                                                        "test_disabled", "kwargs"],
                                    "Branch": ["initial_name", "elements"],
                                    "Stack": ["initial_name", "elements"],
                                    "Switch": ["initial_name", "elements"],
                                    "DataFilter": ["indices"],
                                    "NeuroBranch": ["initial_name", "elements"]
                                    }

    @staticmethod
    def write_json_file(value: dict, path: str):
        """
        static method for dumping dict json
        :param value: dict to dump
        :param path: storage path
        :return: None
        """
        with open(path, 'w') as outfile:
            json.dump(value, outfile, indent=4)
        return

    @staticmethod
    def read_json_file(filepath: str):
        """
        reverse function to write_json_file, read json file -> dict
        :param filepath: storage path
        :return: dict type
        """
        with open(filepath, 'r') as outfile:
            value = json.load(outfile)
        return value

    @staticmethod
    def str_to_class(classname: str):
        """
        function for bringing classname to life
        :param classname: String-name of class
        :return: class
        """

        if classname in [x[0] for x in inspect.getmembers(sys.modules[__name__])]:
            obj = getattr(sys.modules[__name__], classname)
        elif classname in [x[0] for x in inspect.getmembers(sys.modules["photonai.base"])]:
            obj = getattr(sys.modules["photonai.base"], classname)
        else:
            msg = "Json Transformer is not able to initialize the hyperpipe. " \
                  "Class: {} is not defined.".format(classname)
            logger.error(msg)
            raise ValueError(msg)
        return obj

    def to_json_file(self, pipe, path: str):
        """
        main function for saving PHOTON.Hyperpipe -> file.json
        :param pipe:
        :param path: Path to save json.
        :return:
        """
        self.data_json = self.create_json(pipe)
        self.write_json_file(self.data_json, path)

    def create_json(self, pipe):
        """
        function for saving a PHOTON.Hyperpipe to json
        :param pipe: Hyperpipe to transform
        :return: dict representation of hyperpipe
        """
        self.data_json = dict()
        self.data_json["name"] = pipe.name
        for key in ["verbosity", "permutation_id", "cache_folder", "nr_of_processes"]:
            self.data_json[key] = getattr(pipe, key)
        self.data_json["random_seed"] = pipe.random_state

        self.data_json["inner_cv"] = self.transform_elements_recursive(pipe.cross_validation.inner_cv)
        self.data_json["outer_cv"] = self.transform_elements_recursive(pipe.cross_validation.outer_cv)
        for c_key in ["calculate_metrics_across_folds", "use_test_set", "test_size",
                      "calculate_metrics_per_fold"]:
            self.data_json[c_key] = getattr(pipe.cross_validation, c_key)

        self.data_json["performance_constraints"] = self.transform_elements_recursive(
            pipe.optimization.performance_constraints)
        self.data_json["optimizer"] = pipe.optimization.optimizer_input_str
        if pipe.optimization.optimizer_params:
            self.data_json["optimizer_params"] = self.transform_elements_recursive(pipe.optimization.optimizer_params)
        self.data_json["metrics"] = self.transform_elements_recursive(pipe.optimization.metrics)
        self.data_json["best_config_metric"] = pipe.optimization.best_config_metric
        self.data_json["project_folder"] = pipe.project_folder

        if pipe.output_settings:
            self.data_json["output_settings"] = {"mongodb_connect_url": pipe.output_settings.mongodb_connect_url,
                                                 "save_output": pipe.output_settings.save_output,
                                                 "overwrite_results": pipe.output_settings.overwrite_results,
                                                 "user_id": pipe.output_settings.user_id,
                                                 "wizard_object_id": pipe.output_settings.wizard_object_id,
                                                 "wizard_project_name": pipe.output_settings.wizard_project_name,
                                                 "__photon_type": "OutputSettings"}

        if pipe.preprocessing:
            self.data_json["preprocessing"] = {"elements":
                                                   self.transform_elements_recursive(pipe.preprocessing.elements),
                                               "__photon_type": "Preprocessing"}

        self.data_json["elements"] = self.transform_elements_recursive(pipe.elements)

        return self.data_json

    def transform_elements_recursive(self, element):
        """
        recursive functinon for Hyperpipe.elements and all class types.
        :param element: as type dict, list, tuple, str, bool, ...
        :return: json like version
        """
        d = {}
        if element is None:
            return d
        # main dtype
        if any(isinstance(element, t) for t in [int, bool, str, float, np.float]):
            return element
        # dtype == list or dtype == tuple
        if isinstance(element, list) or isinstance(element, tuple):
            tmp_list = []
            for e in element:
                tmp_list.append(self.transform_elements_recursive(e))
            if isinstance(element, list):
                return tmp_list
            else:
                return tuple(tmp_list)
        # dtype == dict
        elif isinstance(element, dict):
            for k, v in element.items():
                if not (k.startswith('_') or v is None):
                    if str(element.__class__.__name__) in self.attribute_allocator.keys():
                        if k in self.attribute_allocator[str(element.__class__.__name__)]:
                            d[k] = self.transform_elements_recursive(v)
                    elif k not in self.black_list:
                        d[k] = self.transform_elements_recursive(v)
        # dtype == object
        else:
            for k, v in (dict(inspect.getmembers(element))).items():
                if not (k.startswith('_') or v is None or inspect.ismethod(v)):
                    if str(element.__class__.__name__) in self.attribute_allocator.keys():
                        if k in self.attribute_allocator[str(element.__class__.__name__)]:
                            d[k] = self.transform_elements_recursive(v)
                    elif k not in self.black_list:
                        d[k] = self.transform_elements_recursive(v)
        if d and str(element.__class__.__name__) not in ["dict", "type", "property"]:
            d['__photon_type'] = str(element.__class__.__name__)
            if d['__photon_type'] in ["IntegerRange", "FloatRange", "NumberType", "BooleanSwitch"] \
                    and 'values' in d:
                del d['values']
        d = {key: val for key, val in d.items() if val is not None}
        if not d:
            return None
        return d

    def from_json_file(self, path: str):
        """
        main function to load from file.json
        :param path: storage path
        :return: json value
        """
        self.data_json = self.read_json_file(path)
        return self.from_json(self.data_json)

    def from_json(self, data_json):
        """
        manual to read hyperpipe json
        :param data_json: storaged json
        :return: PHOTON.Hyperpipe
        """
        self.data_json = data_json
        for key in ['inner_cv', 'outer_cv', 'output_settings', 'performance_constraints']:
            self.data_json[key] = self.load_elements_recursive(self.data_json[key])
        init = {key: value for key, value in self.data_json.items() if key not in ["elements", "preprocessing"]}
        pipe = self.str_to_class("Hyperpipe")(**init)
        elements = self.load_elements_recursive(self.data_json["elements"])
        if "preprocessing" in data_json:
            pre_elements = self.load_elements_recursive(self.data_json["preprocessing"]["elements"])
            pre = Preprocessing()
            for pre_element in pre_elements:
                pre += pre_element
            pipe += pre
        for element in elements:
            pipe += element
        return pipe

    def load_elements_recursive(self, data_json):
        """
        reverse of transform_elements_recursive
        :param data_json: dict, list, tuple, str, bool, ...
        :return: python class/type
        """
        a = {}
        b = []
        if data_json is None:
            return None
        # main dtypes
        if any(isinstance(data_json, t) for t in [int, bool, str, float]):
            return data_json
        # dtype == list
        if isinstance(data_json, list):
            for element in data_json:
                b.append(self.load_elements_recursive(element))
            return b
        # dtype == dict
        elif isinstance(data_json, dict):
            if not data_json:
                return None
            for element in data_json.keys():
                if element not in ["initial_name", "__photon_type", "kwargs", "initial_hyperparameters"]:
                    a.update({element: self.load_elements_recursive(data_json[element])})
                elif element == "kwargs":
                    for key in data_json["kwargs"].keys():
                        a.update({key: self.load_elements_recursive(data_json[element][key])})
                elif element == "initial_hyperparameters":
                    tmp = {}
                    for key in data_json[element].keys():
                        if key == "__photon_type":
                            continue
                        tmp.update({key: self.load_elements_recursive(data_json[element][key])})
                    a["hyperparameters"] = tmp
                elif element == "initial_name":
                    a["name"] = data_json['initial_name']
            if "__photon_type" in a:
                del a["__photon_type"]
            return self.str_to_class(data_json[element])(**a)



import unittest
import numpy as np
from sklearn.datasets import load_breast_cancer

from photonai.processing.cross_validation import StratifiedKFoldRegression
from photonai.processing.photon_folds import FoldInfo


class StratifiedRegressionTest(unittest.TestCase):

    def setUp(self):
        # set up all we need
        self.X, self.y = load_breast_cancer(return_X_y=True)
        self.kwargs = {'groups': np.random.randint(-10, 10, (self.y.shape[0]))}
        self.outer_cv = StratifiedKFoldRegression(n_splits=2, random_state=15, shuffle=True)
        self.inner_cv = StratifiedKFoldRegression(n_splits=2, random_state=15, shuffle=True)
        self.outer_folds = FoldInfo.generate_folds(self.outer_cv, self.X, self.y, {})

        # it's important to say at this point, that "groups" is a giant misnomer for StratifiedKFoldRegression
        # if there is a groups variable in kwargs, PHOTON will forward this variable to StatifiedKFoldRegression
        # however, this variable shouldn't be a group but a continuous variable
        self.outer_folds_w_groups = FoldInfo.generate_folds(self.outer_cv, self.X, self.y, self.kwargs)

    def test_stratified_regression(self):
        splits = list()
        splits_pipe = list()
        for train, test in self.outer_cv.split(self.X, self.y):
            splits.append(train)
            splits.append(test)

        for fold in self.outer_folds:
            splits_pipe.append(fold.train_indices)
            splits_pipe.append(fold.test_indices)

        for i in range(len(splits)):
            np.testing.assert_array_equal(splits[i], splits_pipe[i])

    def test_stratified_regression_w_groups(self):
        splits = list()
        splits_pipe = list()
        for train, test in self.outer_cv.split(self.X, self.kwargs['groups']):
            splits.append(train)
            splits.append(test)

        for fold in self.outer_folds_w_groups:
            splits_pipe.append(fold.train_indices)
            splits_pipe.append(fold.test_indices)

        for i in range(len(splits)):
            np.testing.assert_array_equal(splits[i], splits_pipe[i])

    def test_multi_dimensional_y(self):
        y = np.random.randint(-10, 10, (self.y.shape[0], 2))
        with self.assertRaises(NotImplementedError):
            FoldInfo.generate_folds(self.outer_cv, self.X, y, {})



from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.svm import OneClassSVM
from sklearn.metrics import accuracy_score
import numpy as np


class PhotonOneClassSVM(BaseEstimator, ClassifierMixin):

    def __init__(self, kernel='rbf', nu=0.5):
        self.kernel = kernel
        self.nu = nu
        self.my_svm = None

    def fit(self, X, y=None):
        self.my_svm = OneClassSVM(kernel=self.kernel, nu=self.nu)
        self.my_svm.fit(X[y==1])
        return self

    def predict(self, X):
        return self.my_svm.predict(X)

    def score(self, X, y):
        y_pred = self.predict(X)

        return accuracy_score(y, y_pred)

    def set_params(self, **params):
        if 'kernel' in params:
            self.kernel = params['kernel']
        if 'nu' in params:
            self.nu = params['nu']

        if not self.my_svm:
            self.my_svm = OneClassSVM(kernel=self.kernel, nu=self.nu)
        self.my_svm.set_params(**params)

    def get_params(self, deep=True):
        if not self.my_svm:
            self.my_svm = OneClassSVM(kernel=self.kernel, nu=self.nu)
        svm_dict = self.my_svm.get_params(deep)
        return svm_dict



import random
import unittest

import numpy as np
from sklearn.datasets import load_breast_cancer

from photonai.helper.helper import PhotonDataHelper
from photonai.processing.photon_folds import FoldInfo


class FoldInfoTests(unittest.TestCase):

    def setUp(self):
        dataset = load_breast_cancer()
        self.__X = dataset.data
        self.__y = dataset.target

    def test_class_distribution_info(self):
        unique, counts = np.unique(self.__y, return_counts=True)
        nr_dict = FoldInfo.data_overview(self.__y)
        self.assertEqual(counts[1], nr_dict['1'])


class DataHelperTests(unittest.TestCase):

    def test_data_split_indices(self):
        vals = np.array([-1, -2, -3, -4, -5, -6, -7, -8, -9, -10])
        vals_str = np.array([ascii(i) for i in vals])
        random_features = np.random.randn(10, 20)
        kwargs = {'test': vals,
                  'subtest': vals_str,
                  'random': random_features}
        pick_list = [1, 3, 5]
        splitted_X, splitted_y, splitted_example = PhotonDataHelper.split_data(random_features, vals,
                                                   kwargs, indices=pick_list)
        self.assertTrue(np.array_equal(splitted_X, random_features[pick_list]))
        self.assertTrue(np.array_equal(splitted_y, vals[pick_list]))
        self.assertTrue(np.array_equal(splitted_example['test'], vals[pick_list]))
        self.assertTrue(np.array_equal(splitted_example['subtest'], vals_str[pick_list]))
        self.assertTrue(np.array_equal(splitted_example['random'], random_features[pick_list]))

    def test_split_join_resorting(self):
        X = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        y = np.array([1, 1, 1, 1, 1, 2, 2, 2, 2, 2])
        kwargs = {'test': np.array([-1, -2, -3, -4, -5, -6, -7, -8, -9, -10])}

        X_new, y_new, kwargs_new = list(), list(), dict()

        # first randomly split the data and append them to X_new, y_new, kwargs_new
        idx_list_one, idx_list_two = list(), list()
        for idx in range(len(X)):
            if bool(random.getrandbits(1)):
                idx_list_one.append(idx)
            else:
                idx_list_two.append(idx)

        for ilist in [idx_list_two, idx_list_one]:
            for idx in ilist:

                X_batched, y_batched, kwargs_batched = PhotonDataHelper.split_data(X, y, kwargs, idx, idx)

                # test if batching works
                self.assertEqual(X_batched, X[idx])
                self.assertEqual(y_batched, y[idx])
                self.assertDictEqual(kwargs_batched, {'test': [kwargs['test'][idx]]})

                # then join again
                X_new, y_new, kwargs_new = PhotonDataHelper.join_data(X_new, X_batched, y_new, y_batched, kwargs_new, kwargs_batched)

        # test if joining works
        joined_idx = PhotonDataHelper.stack_data_vertically(idx_list_two, idx_list_one)
        self.assertTrue(np.array_equal(X_new, X[joined_idx]))
        self.assertTrue(np.array_equal(y_new, y[joined_idx]))
        self.assertTrue(np.array_equal(kwargs_new['test'], kwargs['test'][joined_idx]))

        # now resort and see if that works too
        X_resorted, y_resorted, kwargs_resorted = PhotonDataHelper.resort_splitted_data(X_new, y_new, kwargs_new, joined_idx)
        self.assertTrue(np.array_equal(X_resorted, X))
        self.assertTrue(np.array_equal(y_resorted, y))
        self.assertListEqual(list(kwargs_resorted.keys()), list(kwargs.keys()))
        self.assertTrue(np.array_equal(kwargs_resorted['test'], kwargs['test']))

    def test_concatenate_dict(self):
        dict_a = {'variable_one': np.random.randn(10),
                  'variable_two': np.random.randn(15)}
        dict_b = {'variable_one': np.random.randn(20),
                  'variable_two': np.random.randn(20)}
        dict_c = {'variable_one': np.random.randn(10, 10),
                  'variable_two': np.random.randn(15, 15)}
        dict_d = {'variable_one': np.random.randn(20, 10),
                  'variable_two': np.random.randn(20, 15)}
        dict_e = {}

        dict_a_b = PhotonDataHelper.join_dictionaries(dict_a, dict_b)
        dict_c_d = PhotonDataHelper.join_dictionaries(dict_c, dict_d)
        dict_e_a = PhotonDataHelper.join_dictionaries(dict_e, dict_a)
        self.assertEqual(len(dict_a_b['variable_one']), 30)
        self.assertEqual(len(dict_a_b['variable_two']), 35)
        self.assertEqual(dict_c_d['variable_one'].shape, (30, 10))
        self.assertEqual(dict_c_d['variable_two'].shape, (35, 15))
        self.assertEqual(len(dict_e_a['variable_one']), 10)
        self.assertEqual(len(dict_e_a['variable_two']), 15)

    def test_index_dict(self):
        labels = np.asarray([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        dict_a = {'variable_one': np.random.randn(10),
                  'variable_two': np.random.randn(10, 10)}
        dict_a_1 = PhotonDataHelper.index_dict(dict_a, labels == 0)
        dict_a_2 = PhotonDataHelper.index_dict(dict_a, labels == 1)
        self.assertEqual(len(dict_a_1['variable_one']), 5)
        self.assertEqual(dict_a_2['variable_two'].shape, (5, 10))



import numpy as np
from sklearn.datasets import load_boston
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement

X, y = load_boston(return_X_y=True)

my_pipe = Hyperpipe(name='single_outer_pipe',
                    metrics=['mean_absolute_error', 'mean_squared_error', 'pearson_correlation'],
                    best_config_metric='mean_absolute_error',
                    use_test_set=False,
                    inner_cv=KFold(n_splits=10, shuffle=True, random_state=42),
                    verbosity=0,
                    project_folder='./tmp/')


# ADD ELEMENTS TO YOUR PIPELINE
my_pipe += PipelineElement('SimpleImputer', missing_values=np.nan, strategy='median')
my_pipe += PipelineElement('StandardScaler')
my_pipe += PipelineElement('GaussianProcessRegressor')


# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)

# find mean and std of all metrics here
test_metrics = my_pipe.results.best_config.metrics_test
train_metrics = my_pipe.results.best_config.metrics_train



import inspect
import importlib
import json
import os
import sys
from glob import glob

import numpy as np
from sklearn.datasets import load_breast_cancer, load_boston
from shutil import copyfile

from photonai.photonlogger.logger import logger


class PhotonRegistry:
    """
    Helper class to manage the PHOTONAI Element Register.

    Use it to add and remove items into the register.
    You can also retrieve information about items and its hyperparameters.

    Every item in the register is encoded by a string literal
    that points to a python class and its namespace.
    You can access the python class via the string literal.
    The class PhotonElement imports and instantiates the class for you.

    Example:
        ``` python
        import os
        from photonai.base import PhotonRegistry

        # REGISTER ELEMENT saved in folder custom_elements_folder
        base_folder = os.path.dirname(os.path.abspath(__file__))
        custom_elements_folder = os.path.join(base_folder, 'custom_elements')

        registry = PhotonRegistry(custom_elements_folder=custom_elements_folder)
        registry.register(photon_name='MyCustomEstimator',
                          class_str='custom_estimator.CustomEstimator',
                          element_type='Estimator')

        registry.activate()
        registry.info('MyCustomEstimator')

        # get informations of other available elements
        registry.info('SVC')
        ```

    """

    ELEMENT_DICTIONARY = dict()
    PHOTON_REGISTRIES = ['PhotonCore']
    CUSTOM_ELEMENTS_FOLDER = None
    CUSTOM_ELEMENTS = None

    def __init__(self, custom_elements_folder: str = None):
        """
        Initialize the object.

        Parameters:
            custom_elements_folder:
                Path to folder with custom element in it.

        """
        self.current_folder = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        self.module_path = os.path.join(self.current_folder, "modules")
        if not os.path.isdir(self.module_path):
            os.mkdir(self.module_path)

        # update list with available sub_elements
        self._list_available_modules()
        PhotonRegistry.CUSTOM_ELEMENTS_FOLDER = custom_elements_folder
        self._load_custom_folder(custom_elements_folder)

        if len(PhotonRegistry.ELEMENT_DICTIONARY) == 0 or \
                PhotonRegistry.ELEMENT_DICTIONARY == PhotonRegistry.CUSTOM_ELEMENTS:
            PhotonRegistry.ELEMENT_DICTIONARY.update(self.get_package_info())

    def _list_available_modules(self):
        for abs_filename in glob(os.path.join(self.module_path, "*.json")):
            basename = os.path.basename(abs_filename)
            file, ext = os.path.splitext(basename)
            if file not in PhotonRegistry.PHOTON_REGISTRIES:
                PhotonRegistry.PHOTON_REGISTRIES.append(file)

    def add_module(self, path_to_file: str):
        filename = os.path.basename(path_to_file)
        copyfile(path_to_file, os.path.join(self.module_path, filename))
        self._list_available_modules()
        PhotonRegistry.ELEMENT_DICTIONARY = self.get_package_info()

    def delete_module(self, module_name: str):
        PhotonRegistry.PHOTON_REGISTRIES.remove(module_name)
        os.remove(os.path.join(self.module_path, module_name + ".json"))
        PhotonRegistry.ELEMENT_DICTIONARY = self.get_package_info()

    def _load_json(self, photon_package: str) -> dict:
        """
        Load JSON file in which the elements for the PHOTON submodule are stored.

        The JSON files are stored in the framework folder by the name convention 'photon_package.json'

        Parameters:
            photon_package:
              The name of the photonai submodule

        Returns:
            JSON file as dict, file path as str.

        """
        if photon_package == 'CustomElements':
            folder = PhotonRegistry.CUSTOM_ELEMENTS_FOLDER
            if not folder:
                return {}
        elif photon_package == "PhotonCore":
            folder = self.current_folder
        else:
            folder = self.module_path

        file_name = os.path.join(folder, photon_package + '.json')
        file_content = {}

        # Reading json
        try:
            with open(file_name, 'r') as f:
                try:
                    file_content = json.load(f)
                except json.JSONDecodeError as jde:
                    # handle empty file
                    if jde.msg == 'Expecting value':
                        logger.error("Package File " + file_name + " was empty.")
                    else:
                        raise jde
        except FileNotFoundError:
            logger.error("Could not find file for " + photon_package)
        if not file_content:
            file_content = dict()
        return file_content

    def get_package_info(self, photon_package: list = PHOTON_REGISTRIES) -> dict:
        """
        Collect all registered elements from JSON file.

        Parameters:
            photon_package:
                The names of the PHOTONAI submodules for which
                the elements should be retrieved.

        Returns:
            Dict of registered elements.

        """
        class_info = dict()
        for package in photon_package:

            content = self._load_json(package)

            for idx, key in enumerate(content):
                class_path, class_name = os.path.splitext(content[key][0])

                if idx == 0 and package not in ["PhotonCore", "CustomElements"]:
                    # try to import something from module.
                    # if that fails. drop this shit.
                    try:
                        imported_module = importlib.import_module(class_path)
                        desired_class = getattr(imported_module, class_name[1:])
                        custom_element = desired_class()
                    except (AttributeError, ModuleNotFoundError) as e:
                        logger.error(e)
                        logger.error("Could not import from package {}. Deleting json.".format(package))
                        self.delete_module(package)

                class_info[key] = class_path, class_name[1:]
        return class_info

    def _load_custom_folder(self, custom_elements_folder):

        if PhotonRegistry.CUSTOM_ELEMENTS_FOLDER is not None:
            PhotonRegistry.CUSTOM_ELEMENTS_FOLDER = self._check_custom_folder(custom_elements_folder)

            # load custom elements from json
            PhotonRegistry.CUSTOM_ELEMENTS = self._load_json('CustomElements')
            PhotonRegistry.PHOTON_REGISTRIES.append('CustomElements')
            PhotonRegistry.ELEMENT_DICTIONARY.update(PhotonRegistry.CUSTOM_ELEMENTS)

    @staticmethod
    def _check_custom_folder(custom_folder):
        if not os.path.exists(custom_folder):
            logger.info('Creating folder {}'.format(custom_folder))
            os.makedirs(custom_folder)

        custom_file = os.path.join(custom_folder, 'CustomElements.json')
        if not os.path.isfile(custom_file):
            logger.info('Creating CustomElements.json')
            with open(custom_file, 'w') as f:
                json.dump('', f)

        return custom_folder

    def activate(self):
        if not PhotonRegistry.CUSTOM_ELEMENTS_FOLDER:
            raise ValueError("To activate a custom elements folder, specify a folder when instantiating the registry "
                             "module. Example: registry = PhotonRegistry('/MY/CUSTOM/ELEMENTS/FOLDER) "
                             "In case you don't have any custom models, there is no need to activate the registry.")
        if not os.path.exists(PhotonRegistry.CUSTOM_ELEMENTS_FOLDER):
            raise FileNotFoundError("Couldn't find custom elements folder: {}".format(PhotonRegistry.CUSTOM_ELEMENTS_FOLDER))
        if not os.path.isfile(os.path.join(PhotonRegistry.CUSTOM_ELEMENTS_FOLDER, 'CustomElements.json')):
            raise FileNotFoundError("Couldn't find CustomElements.json. Did you register your element first?")

        # add folder to python path
        logger.info("Adding custom elements folder to system path...")
        sys.path.append(PhotonRegistry.CUSTOM_ELEMENTS_FOLDER)

        PhotonRegistry.ELEMENT_DICTIONARY.update(self.get_package_info(['CustomElements']))
        logger.info('Successfully activated custom elements!')

    def register(self, photon_name: str, class_str: str, element_type: str):
        """
        Save element information to the JSON file.

        Parameters:
            photon_name:
                The string literal with which you want to access the class.

            class_str:
                The namespace of the class, like in the import statement.

            element_type:
                Can be 'Estimator' or 'Transformer'

        """
        # check if folder exists
        if not PhotonRegistry.CUSTOM_ELEMENTS_FOLDER:
            raise ValueError("To register an element, specify a custom elements folder when instantiating the registry "
                             "module. Example: registry = PhotonRegistry('/MY/CUSTOM/ELEMENTS/FOLDER)")

        if not element_type == "Estimator" and not element_type == "Transformer":
            raise ValueError("Variable element_type must be 'Estimator' or 'Transformer'")

        duplicate = self._check_duplicate(photon_name=photon_name, class_str=class_str, content=PhotonRegistry.CUSTOM_ELEMENTS)

        if not duplicate:
            python_file = os.path.join(PhotonRegistry.CUSTOM_ELEMENTS_FOLDER, class_str.split('.')[0] + '.py')
            if not os.path.isfile(python_file):
                raise FileNotFoundError("Couldn't find python file {} in your custom elements folder. "
                                        "Please copy your file into this folder first!".format(python_file))
            # add new element
            PhotonRegistry.CUSTOM_ELEMENTS[photon_name] = class_str, element_type

            # write back to file
            self._write_to_json(PhotonRegistry.CUSTOM_ELEMENTS)
            logger.info('Adding PipelineElement ' + class_str + ' to CustomElements.json as "' + photon_name + '".')

            # activate custom elements
            self.activate()

            # check custom element
            logger.info("Running tests on custom element...")
            return self._run_tests(photon_name, element_type)
        else:
            logger.error('Could not register element!')

    def _run_tests(self, photon_name, element_type):

        # this is a sneaky hack to avoid circular imports of PipelineElement
        imported_module = importlib.import_module("photonai.base")
        desired_class = getattr(imported_module, "PipelineElement")
        custom_element = desired_class(photon_name)

        # check if has fit, transform, predict
        if not hasattr(custom_element.base_element, 'fit'):
            raise NotImplementedError("Custom element does not implement fit() method.")

        if element_type == 'Transformer' and not hasattr(custom_element.base_element, 'transform'):
            raise NotImplementedError("Custom element does not implement transform() method.")

        if element_type == 'Estimator' and not hasattr(custom_element.base_element, 'predict'):
            raise NotImplementedError("Custom element does not implement predict() method.")

        # check if estimator is regressor or classifier
        if element_type == 'Estimator':
            if hasattr(custom_element, '_estimator_type'):
                est_type = getattr(custom_element, '_estimator_type')
                if est_type == "regressor":
                    X, y = load_boston(return_X_y=True)
                elif est_type == "classifier":
                    X, y = load_breast_cancer(return_X_y=True)
                else:
                    raise ValueError("Custom element does not specify whether it is a regressor or classifier. "
                                     "Is {}".format(est_type))
            else:
                raise NotImplementedError("Custom element does not specify whether it is a regressor or classifier. "
                                          "Consider inheritance from ClassifierMixin or RegressorMixin or set "
                                          "_estimator_type manually.")
        else:
            X, y = load_boston(return_X_y=True)

        # try and test functionality
        kwargs = {'covariates': np.random.randn(len(y))}

        try:
            # test fit
            returned_element = custom_element.base_element.fit(X, y, **kwargs)
        except Exception as e:
            logger.info("Not able to run tests on fit() method. Test data not compatible.")
            return e

        if not isinstance(returned_element, custom_element.base_element.__class__):
            raise NotImplementedError("fit() method does not return self.")

        try:
            # test transform or predict (if base element does not implement transform method, predict should be called
            # by PipelineElement -> so we only need to test transform()
            if custom_element.needs_y:
                if element_type == 'Estimator':
                    raise NotImplementedError("Estimator should not need y.")
                Xt, yt, kwargst = custom_element.base_element.transform(X, y, **kwargs)
                if 'covariates' not in kwargst.keys():
                    raise ValueError("Custom element does not correctly transform kwargs although needs_y is True. "
                                     "If you change the number of samples in transform(), make sure to transform kwargs "
                                     "respectively.")
                if not len(kwargst['covariates']) == len(X):
                    raise ValueError("Custom element is not returning the correct number of samples!")

            elif custom_element.needs_covariates:
                if element_type == 'Estimator':
                    yt, kwargst = custom_element.base_element.predict(X, **kwargs)
                    if not len(yt) == len(y):
                        raise ValueError("Custom element is not returning the correct number of samples!")
                else:
                    Xt, kwargst = custom_element.base_element.transform(X, **kwargs)

                    if not len(Xt) == len(X) or not len(kwargst['covariates']) == len(X):
                        raise ValueError("Custom element is not returning the correct number of samples!")

            else:
                if element_type == 'Estimator':
                    yt = custom_element.base_element.predict(X)
                    if not len(yt) == len(y):
                        raise ValueError("Custom element is not returning the correct number of samples!")
                else:
                    Xt = custom_element.base_element.transform(X)
                    if not len(Xt) == len(X):
                        raise ValueError("Custom element is not returning the correct number of samples!")

        except ValueError as ve:
            if "too many values to unpack" in ve.args[0]:
                raise ValueError("Custom element does not return X, y and kwargs the way it should "
                                          "according to needs_y and needs_covariates.")
            else:
                logger.info(ve.args[0])
                return ve
        except Exception as e:
            logger.info(e.args[0])
            logger.info("Not able to run tests on transform() or predict() method. Test data probably not compatible.")
            return e

        logger.info('All tests on custom element passed.')

    def info(self, photon_name: str):
        """
        Show information for object that is encoded by this name.

        Parameters:
            photon_name:
                The string literal which accesses the class.

        """
        content = self.get_package_info()  # load existing json

        if photon_name in content:
            element_namespace, element_name = content[photon_name]

            print("----------------------------------")
            print("Name: " + element_name)
            print("Namespace: " + element_namespace)
            print("----------------------------------")

            try:
                imported_module = __import__(element_namespace, globals(), locals(), element_name, 0)
                desired_class = getattr(imported_module, element_name)
                base_element = desired_class()
                print("Possible Hyperparameters as derived from constructor:")
                class_args = inspect.signature(base_element.__init__)
                for item, more_info in class_args.parameters.items():
                    print("{:<35} {:<75}".format(item, str(more_info)))
                print("----------------------------------")
            except Exception as e:
                logger.error(e)
                logger.error("Could not instantiate class " + element_namespace + "." + element_name)
        else:
            logger.error("Could not find element " + photon_name)

    def delete(self, photon_name: str):
        """
        Delete Element from JSON file.

        Parameters:
            photon_name:
                The string literal encoding the class.

        """
        if photon_name in PhotonRegistry.CUSTOM_ELEMENTS:
            del PhotonRegistry.CUSTOM_ELEMENTS[photon_name]

            self._write_to_json(PhotonRegistry.CUSTOM_ELEMENTS)
            logger.info('Removing the PipelineElement named "{0}" from CustomElements.json.'.format(photon_name))
        else:
            logger.info('Cannot remove "{0}" from CustomElements.json. Element has not been registered before.'.format(photon_name))

    @staticmethod
    def _check_duplicate(photon_name, class_str: str, content: str) -> bool:
        """
        Helper function to check if the entry is either registered by
        a different name or if the name is already given to another class.

         Parameters:
            content:
                The content of the CustomElements.json.

            class_str:
                The namespace.Classname, where the class lives,
                from where it should be imported.

            photon_name:
                The name of the element with which it is called within PHOTONAI.

        Returns:
             False if there is no key with this name and the class
             is not already registered with another key.

        """
        # check for duplicate name (dict key)
        if photon_name in content:
            logger.info('A PipelineElement named ' + photon_name + ' has already been registered.')
            return True

        # check for duplicate class_str
        if any(class_str in '.'.join([s[0], s[1]]) for s in content.values()):
            logger.info('The Class named ' + class_str + ' has already been registered.')
            return True
        return False

    def _write_to_json(self, content_to_write: dict):
        """
        Write json content to file

        Parameters:
            content2write:
                The new information to attach to the file.

        """
        # Writing JSON data
        with open(os.path.join(self.CUSTOM_ELEMENTS_FOLDER, "CustomElements.json"), 'w') as f:
            json.dump(content_to_write, f)

    def list_available_elements(self, photon_package: list = PHOTON_REGISTRIES):
        """
        Print info about all items that are registered for the PHOTONAI
        submodule to the console.

        Parameters:
            photon_package:
                The names of the PHOTON submodules for which
                the elements should be retrieved.

        """
        if isinstance(photon_package, str):
            photon_package = [photon_package]
        for package in photon_package:
            content = self._load_json(package)
            if len(content) > 0:
                print('\n' + package)
                for k, v in sorted(content.items()):
                    class_info, package_type = v
                    print("{:<35} {:<75} {:<5}".format(k, class_info, package_type))



import numpy as np
from photonai.optimization import GridSearchOptimizer, RandomGridSearchOptimizer, \
    SkOptOptimizer, RandomSearchOptimizer, SMACOptimizer, NevergradOptimizer
from photonai.optimization.switch_optimizer.meta_optimizer import MetaHPOptimizer
from photonai.processing.metrics import Scorer
from photonai.photonlogger.logger import logger


class Optimization:

    OPTIMIZER_DICTIONARY = {'grid_search': GridSearchOptimizer,
                            'random_grid_search': RandomGridSearchOptimizer,
                            'sk_opt': SkOptOptimizer,
                            'smac': SMACOptimizer,
                            'random_search': RandomSearchOptimizer,
                            'nevergrad': NevergradOptimizer,
                            'switch': MetaHPOptimizer}

    def __init__(self, optimizer_input, optimizer_params,
                 metrics, best_config_metric, performance_constraints):

        self._optimizer_input = ''
        self.optimizer_input_str = optimizer_input
        self.optimizer_params = optimizer_params
        self._best_config_metric = ''
        self.performance_constraints = performance_constraints
        self.metrics = None
        self.maximize_metric = None
        self.best_config_metric = None
        self.sanity_check_metrics(metrics, best_config_metric)

    @property
    def best_config_metric(self):
        return self._best_config_metric

    @best_config_metric.setter
    def best_config_metric(self, value):
        self._best_config_metric = value
        if isinstance(self.best_config_metric, str):
            self.maximize_metric = Scorer.greater_is_better_distinction(self.best_config_metric)

    @property
    def optimizer_input_str(self):
        return self._optimizer_input

    @optimizer_input_str.setter
    def optimizer_input_str(self, value):
        if isinstance(value, str):
            if value not in self.OPTIMIZER_DICTIONARY:
                raise ValueError("Optimizer " + value + " not supported right now.")
        self._optimizer_input = value

    def sanity_check_metrics(self, metrics, best_config_metric):

        # first of all register all custom elements, if any
        if best_config_metric is not None and not isinstance(best_config_metric, str) \
                and not isinstance(best_config_metric, list):
            best_config_metric = Scorer.register_custom_metric(best_config_metric)

        if metrics is not None and len(metrics) > 0:
            for i in range(len(metrics)):
                if not isinstance(metrics[i], str):
                    metrics[i] = Scorer.register_custom_metric(metrics[i])

        if metrics is not None:
            metrics = list(filter(None, metrics))
        self.metrics = metrics
        self.best_config_metric = best_config_metric

        if self.metrics is None or len(self.metrics) == 0:
            if self.best_config_metric is None:
                error_msg = "No metrics were chosen. Please choose metrics to quantify your performance and set " \
                            "the best_config_metric so that PHOTON which optimizes for"
                logger.error(error_msg)
                raise ValueError(error_msg)
            else:
                # if only best_config_metric is given, copy if to list of metrics
                self.metrics = [self.best_config_metric]

        if self.best_config_metric is not None:
            if isinstance(self.best_config_metric, list):
                warning_text = "Best Config Metric must be a single metric given as string, no list. " \
                               "PHOTON chose the first one from the list of metrics to calculate."

                self.best_config_metric = self.best_config_metric[0]
                logger.warning(warning_text)
                raise Warning(warning_text)

            # if best_config_metric is not given in metrics list, copy it to list
            if self.best_config_metric not in self.metrics:
                self.metrics.append(self.best_config_metric)

        if self.best_config_metric is None and self.metrics is not None and len(self.metrics) > 0:
            self.best_config_metric = self.metrics[0]
            warning_text = "No best config metric was given, so PHOTON chose the first in the list of metrics as " \
                           "criteria for choosing the best configuration."
            logger.warning(warning_text)
            raise Warning(warning_text)

    def get_optimizer(self):
        if isinstance(self.optimizer_input_str, str):
            # instantiate optimizer from string
            optimizer_class = self.OPTIMIZER_DICTIONARY[self.optimizer_input_str]
            optimizer_instance = optimizer_class(**self.optimizer_params)
            return optimizer_instance
        else:
            # Todo: check if object has the right interface
            return self.optimizer_input_str

    def get_optimum_config_outer_folds(self, outer_folds):
        list_of_scores = list()
        for outer_fold in outer_folds:
            metrics = outer_fold.best_config.best_config_score.validation.metrics
            list_of_scores.append(metrics[self.best_config_metric])

        if self.maximize_metric:
            # max metric
            best_config_metric_nr = np.argmax(list_of_scores)
        else:
            # min metric
            best_config_metric_nr = np.argmin(list_of_scores)

        best_config = outer_folds[best_config_metric_nr].best_config
        return best_config


import pickle
import uuid
from enum import Enum

import numpy as np
from pymodm import MongoModel, EmbeddedMongoModel, fields


class MetricHelper:

    def get_train_metric(self, name="", operation="mean"):
        return MDBHelper.get_metric(self.metrics_train, name, operation)

    def get_test_metric(self, name="", operation="mean"):
        return MDBHelper.get_metric(self.metrics_test, name, operation)

    def get_train_metric_dict(self):
        return MDBHelper.get_dict_from_metric_list(self.metrics_train)

    def get_test_metric_dict(self):
        return MDBHelper.get_dict_from_metric_list(self.metrics_test)


class MDBScoreInformation(EmbeddedMongoModel):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    metrics = fields.DictField(blank=True)
    score_duration = fields.IntegerField(blank=True)
    y_true = fields.ListField(blank=True)
    y_pred = fields.ListField(blank=True)
    indices = fields.ListField(blank=True)
    probabilities = fields.ListField(blank=True)
    metrics_copied_from_inner = fields.BooleanField(default=False)

    def save_memory(self):
        self.y_true = list()
        self.y_pred = list()
        self.indices = list()
        self.probabilities = list()

    def __str__(self):
        return str(self.metrics)


class MDBInnerFold(EmbeddedMongoModel):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    fold_nr = fields.IntegerField()
    training = fields.EmbeddedDocumentField(MDBScoreInformation, blank=True)
    validation = fields.EmbeddedDocumentField(MDBScoreInformation, blank=True)
    number_samples_training = fields.IntegerField(blank=True)
    number_samples_validation = fields.IntegerField(blank=True)
    time_monitor = fields.DictField(blank=True)
    feature_importances = fields.ListField(blank=True)
    learning_curves = fields.ListField(blank=True)


class MDBFoldMetric(EmbeddedMongoModel):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    operation = fields.CharField(blank=True, default='')
    metric_name = fields.CharField(blank=True)
    value = fields.FloatField(blank=True)

    def __str__(self):
        return "__".join([self.metric_name, self.operation, str(self.value)])


class MDBConfig(EmbeddedMongoModel, MetricHelper):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    photon_config_id = fields.CharField(blank=True)
    inner_folds = fields.EmbeddedDocumentListField(MDBInnerFold, default=[], blank=True)
    best_config_score = fields.EmbeddedDocumentField(MDBInnerFold, blank=True)
    computation_start_time = fields.DateTimeField(blank=True)
    computation_end_time = fields.DateTimeField(blank=True)
    fit_duration_minutes = fields.IntegerField(blank=True)
    pipe_name = fields.CharField(blank=True)
    config_dict = fields.DictField(blank=True)
    children_config_dict = fields.DictField(blank=True)
    children_config_ref = fields.ListField(default=[], blank=True)
    # best_config_ref_to_train_item = fields.CharField(blank=True)
    config_nr = fields.IntegerField(blank=True)
    config_failed = fields.BooleanField(default=False)
    config_error = fields.CharField(blank=True)
    full_model_spec = fields.DictField(blank=True)
    metrics_train = fields.EmbeddedDocumentListField(MDBFoldMetric, default=[], blank=True)
    metrics_test = fields.EmbeddedDocumentListField(MDBFoldMetric, default=[], blank=True)
    human_readable_config = fields.DictField(blank=True)

    def set_photon_id(self):
        self.photon_config_id = str(uuid.uuid4())

    def decrease_memory(self):
        for fold in self.inner_folds:
            fold.training.save_memory()
            fold.validation.save_memory()
            fold.feature_importances = []
            fold.time_monitor = {}


class MDBOuterFold(EmbeddedMongoModel):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    fold_nr = fields.IntegerField(blank=True)
    best_config = fields.EmbeddedDocumentField(MDBConfig, blank=True)
    tested_config_list = fields.EmbeddedDocumentListField(MDBConfig, default=[], blank=True)
    number_samples_test = fields.IntegerField(blank=True)
    class_distribution_test = fields.DictField(blank=True, default={})
    class_distribution_validation = fields.DictField(blank=True, default={})
    number_samples_validation = fields.IntegerField(blank=True)
    dummy_results = fields.EmbeddedDocumentField(MDBInnerFold, blank=True)

    def get_optimum_config(self, metric, maximize_metric, dict_filter=None, fold_operation="mean"):
        """
        Looks for the best configuration according to the metric with which the configurations are compared
        :param tested_configs: the list of tested configurations and their performances
        :return: MDBConfiguration that has performed best
        """

        list_of_non_failed_configs = [conf for conf in self.tested_config_list if not conf.config_failed]

        # filter configs by key value pair (e.g. estimators__estimator == "SVC")
        if dict_filter is not None:
            list_of_non_failed_configs = [conf for conf in list_of_non_failed_configs
                                          if dict_filter[0] in conf.config_dict
                                          and conf.config_dict[dict_filter[0]] == dict_filter[1]]

        if len(list_of_non_failed_configs) == 0:
            raise Warning("No Configs found which did not fail.")

        if len(list_of_non_failed_configs) == 1:
            best_config_outer_fold = list_of_non_failed_configs[0]
        else:
            list_of_config_vals = [c.get_test_metric(metric, fold_operation) for c in list_of_non_failed_configs]

            if maximize_metric:
                # max metric
                best_config_metric_nr = np.argmax(list_of_config_vals)
            else:
                # min metric
                best_config_metric_nr = np.argmin(list_of_config_vals)

            best_config_outer_fold = list_of_non_failed_configs[best_config_metric_nr]

        return best_config_outer_fold


class MDBPermutationMetrics(EmbeddedMongoModel):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    metric_name = fields.CharField(blank=True)
    metric_value = fields.FloatField(blank=True)
    p_value = fields.FloatField(blank=True)
    values_permutations = fields.ListField(blank=True)


class MDBPermutationResults(EmbeddedMongoModel):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    n_perms = fields.IntegerField(blank=True)
    n_perms_done = fields.IntegerField(blank=True)
    random_state = fields.IntegerField(blank=True)
    metrics = fields.EmbeddedDocumentListField(MDBPermutationMetrics, blank=True)


class MDBDummyResults(EmbeddedMongoModel, MetricHelper):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    strategy = fields.CharField(blank=True)
    metrics_train = fields.EmbeddedDocumentListField(MDBFoldMetric, default=[], blank=True)
    metrics_test = fields.EmbeddedDocumentListField(MDBFoldMetric, default=[], blank=True)


class MDBHyperpipeInfo(EmbeddedMongoModel):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    data = fields.DictField(blank=True)
    cross_validation = fields.DictField(blank=True)
    optimization = fields.DictField(blank=True)
    elements = fields.DictField(blank=True)
    metrics = fields.ListField(blank=True)
    best_config_metric = fields.CharField(blank=True)
    maximize_best_config_metric = fields.BooleanField(blank=True)
    estimation_type = fields.CharField(blank=True)
    eval_final_performance = fields.BooleanField(blank=True)
    # todo: deprecated!!! delete in later versions.
    flowchart = fields.CharField(blank=True)


class MDBHyperpipe(MongoModel, MetricHelper):
    class Meta:
        final = True
        connection_alias = 'photon_core'

    name = fields.CharField()
    version = fields.CharField()
    output_folder = fields.CharField(blank=True)

    permutation_id = fields.CharField(blank=True)
    permutation_failed = fields.CharField(blank=True)
    permutation_test = fields.EmbeddedDocumentField(MDBPermutationResults, blank=True)

    computation_completed = fields.BooleanField(default=False)
    computation_start_time = fields.DateTimeField(blank=True)
    computation_end_time = fields.DateTimeField(blank=True)

    outer_folds = fields.EmbeddedDocumentListField(MDBOuterFold, default=[], blank=True)
    best_config = fields.EmbeddedDocumentField(MDBConfig, blank=True)
    best_config_feature_importances = fields.ListField(blank=True)
    metrics_train = fields.EmbeddedDocumentListField(MDBFoldMetric, default=[], blank=True)
    metrics_test = fields.EmbeddedDocumentListField(MDBFoldMetric, default=[], blank=True)

    hyperpipe_info = fields.EmbeddedDocumentField(MDBHyperpipeInfo)

    # dummy estimator
    dummy_estimator = fields.EmbeddedDocumentField(MDBDummyResults, blank=True)

    # stuff for wizard connection
    user_id = fields.CharField(blank=True)
    wizard_object_id = fields.ObjectIdField(blank=True)
    wizard_system_name = fields.CharField(blank=True)


class ParallelData(MongoModel):

    unprocessed_data = fields.ObjectIdField()
    processed_data = fields.ObjectIdField()


class MDBHelper:
    OPERATION_DICT = {"mean": np.mean, "std": np.std}

    @staticmethod
    def get_metric(metric_list, name="", operation=""):
        if name and operation:
            metric = [i for i in metric_list if i.metric_name == name and i.operation == operation]
            if len(metric) == 0:
                return None
            if len(metric) == 1:
                return metric[0].value
            else:
                raise KeyError("Found multiple metrics with same operation and name.")
        elif not name and operation:
            metric_list = {i.metric_name: i.value for i in metric_list if i.operation == operation}
            return metric_list
        else:
            raise ValueError("Can get metric(s) either by passing name and operation (raw, mean, std),"
                             "or operation only. No arguments are given. ")
            return None

    @staticmethod
    def get_dict_from_metric_list(metric_list):
        best_config_metrics = {}
        for train_metric in metric_list:
            if train_metric.metric_name not in best_config_metrics:
                best_config_metrics[train_metric.metric_name] = {}
            best_config_metrics[train_metric.metric_name][train_metric.operation] = np.round(train_metric.value, 6)
        return best_config_metrics

    @staticmethod
    def aggregate_metrics_for_outer_folds(outer_folds, metrics):
        folds = [fold.best_config.best_config_score for fold in outer_folds]
        return MDBHelper._aggregate_metrics(folds, metrics)

    @staticmethod
    def aggregate_metrics_for_inner_folds(inner_folds, metrics):
        return MDBHelper._aggregate_metrics(inner_folds, metrics)

    @staticmethod
    def _aggregate_metrics(folds, metrics):

        def calculate_single_metric(operation_name, value_list: list, **kwargs):
            if operation_name in MDBHelper.OPERATION_DICT:
                val = MDBHelper.OPERATION_DICT[operation_name](value_list, **kwargs)
            else:
                raise KeyError('Could not find function for processing metrics across folds:' + operation_name)
            return val

        operations = MDBHelper.OPERATION_DICT.keys()
        metrics_train = []
        metrics_test = []
        for metric_item in metrics:
            for op in operations:
                value_list_train = [fold.training.metrics[metric_item] for fold in folds
                                    if fold.training is not None and metric_item in fold.training.metrics]
                if value_list_train:
                    metrics_train.append(MDBFoldMetric(operation=op, metric_name=metric_item,
                                                       value=calculate_single_metric(op, value_list_train)))

                value_list_validation = [fold.validation.metrics[metric_item] for fold in folds
                                         if fold.validation is not None and metric_item in fold.validation.metrics]
                if value_list_validation:
                    metrics_test.append(MDBFoldMetric(operation=op, metric_name=metric_item,
                                                      value=calculate_single_metric(op, value_list_validation)))
        return metrics_train, metrics_test

    @staticmethod
    def load_results(filename):
        return MDBHyperpipe.from_document(pickle.load(open(filename, 'rb')))



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement, Stack, Branch, Switch, DataFilter
from photonai.optimization import FloatRange, IntegerRange

# LOAD DATA FROM SKLEARN
X, y = load_breast_cancer(return_X_y=True)

my_pipe = Hyperpipe('data_integration',
                    optimizer='random_grid_search',
                    optimizer_params={'n_configurations': 20},
                    metrics=['accuracy', 'precision', 'recall'],
                    best_config_metric='f1_score',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=3),
                    verbosity=0,
                    project_folder='./tmp/')

my_pipe += PipelineElement('SimpleImputer')
my_pipe += PipelineElement('StandardScaler', {}, with_mean=True)

# Use only "mean" features: [mean_radius, mean_texture, mean_perimeter, mean_area, mean_smoothness, mean_compactness,
# mean_concavity, mean_concave_points, mean_symmetry, mean_fractal_dimension
mean_branch = Branch('MeanFeature')
mean_branch += DataFilter(indices=range(0, 10))
mean_branch += PipelineElement('SVC', {'C': FloatRange(0.1, 200)}, kernel='linear')

# Use only "error" features
error_branch = Branch('ErrorFeature')
error_branch += DataFilter(indices=range(10, 20))
error_branch += PipelineElement('SVC', {'C': FloatRange(0.1, 200)}, kernel='linear')

# use only "worst" features: [worst_radius, worst_texture, ..., worst_fractal_dimension]
worst_branch = Branch('WorstFeature')
worst_branch += DataFilter(indices=range(20, 30))
worst_branch += PipelineElement('SVC', {'C': FloatRange(0.1, 200)}, kernel='linear')

my_pipe += Stack('SourceStack', [mean_branch, error_branch, worst_branch])

my_pipe += Switch('EstimatorSwitch', [PipelineElement('RandomForestClassifier', {'n_estimators': IntegerRange(2, 5)}),
                                      PipelineElement('SVC')])

my_pipe.fit(X, y)



from sklearn.model_selection import KFold
from sklearn.datasets import load_breast_cancer

from photonai.base import Hyperpipe, Switch, Stack, Branch, PipelineElement
from photonai.optimization import IntegerRange

X, y = load_breast_cancer(return_X_y=True)

my_pipe = Hyperpipe('voting_example',
                    inner_cv=KFold(n_splits=10),
                    outer_cv=KFold(n_splits=5),
                    metrics=['balanced_accuracy', 'precision'],
                    best_config_metric='balanced_accuracy',
                    project_folder='./tmp')

my_pipe += PipelineElement("SimpleImputer")

switch = Switch("my_copy_switch")
switch += PipelineElement("StandardScaler")
switch += PipelineElement("RobustScaler", test_disabled=True)
my_pipe += switch

stack = Stack("EstimatorStack")

branch1 = Branch("SVMBranch")
branch1 += PipelineElement("PCA", hyperparameters={'n_components': IntegerRange(5, 10)})
branch1 += PipelineElement("SVC")

branch2 = Branch('TreeBranch')
branch2 += PipelineElement("PCA", hyperparameters={'n_components': IntegerRange(5, 10)})
branch2 += PipelineElement("DecisionTreeClassifier")

stack += branch1
stack += branch2
my_pipe += stack

my_pipe += PipelineElement("PhotonVotingClassifier")

my_pipe.fit(X, y)



import pandas as pd
from sklearn.model_selection import KFold, ShuffleSplit

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import IntegerRange, MinimumPerformanceConstraint

# setup training and test workflow
my_pipe = Hyperpipe('heart_failure_final_perf',
                    outer_cv=ShuffleSplit(n_splits=100, test_size=0.2),
                    inner_cv=KFold(n_splits=10, shuffle=True),
                    use_test_set=True,
                    metrics=['balanced_accuracy', 'f1_score', 'matthews_corrcoef',
                             'sensitivity', 'specificity'],
                    best_config_metric='f1_score',
                    optimizer='grid_search',
                    project_folder='./tmpv2',
                    performance_constraints=MinimumPerformanceConstraint('f1_score', threshold=0.7, strategy="mean"),
                    cache_folder='./cache',
                    verbosity=0)


# arrange a sequence of algorithms subsequently applied
my_pipe += PipelineElement('StandardScaler')

my_pipe += PipelineElement('SimpleImputer')


my_pipe += PipelineElement('ImbalancedDataTransformer',
                           hyperparameters={'method_name': ['RandomUnderSampler', 'RandomOverSampler', 'SMOTE']},
                           test_disabled=True)

# compare different learning algorithms in an OR_Element
my_pipe += PipelineElement('RandomForestClassifier', criterion='gini', bootstrap=True,
                           hyperparameters={'min_samples_split': IntegerRange(2, 30),
                                            'max_features': ['auto', 'sqrt', 'log2']})

# read data
df = pd.read_csv('./heart_failure_clinical_records_dataset.csv')
X = df.iloc[:, 0:12]
y = df.iloc[:, 12]

# start the training, optimization and test procedure
my_pipe.fit(X, y)

debug = True



from sklearn.datasets import load_breast_cancer, load_boston
from keras.models import Sequential
from keras.layers import Dense, Dropout
import tensorflow as tf
import numpy as np
import warnings
import os

import unittest
from photonai.modelwrapper.keras_base_models import KerasBaseClassifier, KerasBaseRegressor


class KerasBaseClassifierTest(unittest.TestCase):

    def setUp(self):
        self.X, self.y = load_breast_cancer(return_X_y=True)

        self.model = Sequential()
        self.model.add(Dense(3, input_dim=self.X.shape[1], activation='relu'))
        self.model.add(Dropout(0.1))
        self.model.add(Dense(2, activation='softmax'))
        self.model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])

        self.estimator_type = KerasBaseClassifier

        inputs = tf.keras.Input(shape=(self.X.shape[1],))
        x = tf.keras.layers.Dense(4, activation=tf.nn.relu)(inputs)
        outputs = tf.keras.layers.Dense(2, activation=tf.nn.softmax)(x)
        self.tf_model = tf.keras.Model(inputs=inputs, outputs=outputs)
        self.tf_model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])

    def test_predict(self):
        estimator = self.estimator_type(model=self.model,
                                        epochs=3,
                                        callbacks=[tf.keras.callbacks.EarlyStopping(patience=2)],
                                        nn_batch_size=2,
                                        verbosity=0
                                        )
        estimator.fit(self.X, self.y)
        res = estimator.predict(self.X)
        self.assertTrue(isinstance(res, np.ndarray))

        # small data input
        with warnings.catch_warnings(record=True) as w:
            estimator.fit(self.X[:99, :], self.y[:99])
            res = estimator.predict(self.X[:20])
            assert any("Cannot use validation split because of small" in s for s in [e.message.args[0] for e in w])
            self.assertTrue(isinstance(res, np.ndarray))

    def test_tf_model(self):
        estimator = self.estimator_type(model=self.tf_model,
                                        epochs=3,
                                        nn_batch_size=2,
                                        verbosity=0
                                        )
        estimator.fit(self.X, self.y)

        estimator.save("keras_example_saved_model")

        reload_estinator = self.estimator_type()
        reload_estinator.load("keras_example_saved_model")

        np.testing.assert_array_almost_equal(estimator.predict(self.X), reload_estinator.predict(self.X), decimal=4)

        # remove saved keras files
        for fname in os.listdir("."):
            if fname.startswith("keras_example_saved_model"):
                os.remove(os.path.join(".", fname))


class KerasBaseRegressorTest(KerasBaseClassifierTest):

    def setUp(self):
        self.X, self.y = load_boston(return_X_y=True)

        self.model = Sequential()
        self.model.add(Dense(5, input_dim=self.X.shape[1], activation='relu'))
        self.model.add(Dropout(0.5))
        self.model.add(Dense(1, activation='linear'))
        self.model.compile(loss='mean_squared_error', optimizer='adam', metrics=['mean_squared_error'])

        self.estimator_type = KerasBaseRegressor

        inputs = tf.keras.Input(shape=(self.X.shape[1],))
        x = tf.keras.layers.Dense(4, activation=tf.nn.relu)(inputs)
        outputs = tf.keras.layers.Dense(1, activation=tf.keras.activations.linear)(x)
        self.tf_model = tf.keras.Model(inputs=inputs, outputs=outputs)
        self.tf_model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])



import os
from sklearn.model_selection import KFold, ShuffleSplit
from sklearn.datasets import load_boston

from photonai.base import Hyperpipe, PipelineElement, Switch
from photonai.optimization import FloatRange, IntegerRange
from photonai.processing import ResultsHandler

X, y = load_boston(return_X_y=True)

# setup training and test workflow
my_pipe = Hyperpipe('compare_estimators',
                    outer_cv=ShuffleSplit(n_splits=5, test_size=0.2),
                    inner_cv=KFold(n_splits=5, shuffle=True),
                    metrics=['mean_absolute_error', 'mean_squared_error', 'explained_variance'],
                    best_config_metric='mean_squared_error',
                    optimizer='switch',
                    optimizer_params={'name': 'sk_opt', 'n_configurations': 10},
                    project_folder='./tmp',
                    verbosity=1)

# arrange a sequence of algorithms subsequently applied
my_pipe += PipelineElement('StandardScaler')

# compare different learning algorithms in an OR_Element
estimators = Switch('estimator_selection')

estimators += PipelineElement('RandomForestRegressor', bootstrap=True,
                              hyperparameters={'min_samples_split': IntegerRange(2, 4),
                                               'max_features': ['auto', 'sqrt', 'log2']})

estimators += PipelineElement('GradientBoostingRegressor',
                              hyperparameters={'learning_rate': FloatRange(0.001, 1, "logspace")})

estimators += PipelineElement('SVR',
                              hyperparameters={'C': FloatRange(0.5, 25),
                                               'kernel': ['linear', 'rbf']})

my_pipe += estimators

# start the training, optimization and test procedure
my_pipe.fit(X, y)

my_pipe.results_handler.get_mean_of_best_validation_configs_per_estimator()

# or after training
results_folder = my_pipe.results_handler.output_settings.results_folder

res = ResultsHandler()
res.load_from_file(os.path.join(results_folder, "photon_result_file.json"))
estimator_performances = res.get_mean_of_best_validation_configs_per_estimator()



import datetime
import numpy as np
from typing import Union, Generator
import sklearn
import warnings

from skopt import Optimizer
from skopt.space import Real, Integer, Dimension
from skopt.space import Categorical as skoptCategorical

from photonai.optimization import FloatRange, IntegerRange, BooleanSwitch, PhotonHyperparam
from photonai.optimization.base_optimizer import PhotonSlaveOptimizer
from photonai.optimization import Categorical as PhotonCategorical
from photonai.photonlogger.logger import logger


class SkOptOptimizer(PhotonSlaveOptimizer):
    """Wrapper for Scikit-Optimize with PHOTONAI.

    Scikit-Optimize, or skopt, is a simple and efficient library to
    minimize (very) expensive and noisy black-box functions.
    It implements several methods for sequential model-based optimization.
    skopt aims to be accessible and easy to use in many contexts.


    Scikit-optimize [usage and implementation details](https://scikit-optimize.github.io/stable/)

    A detailed parameter documentation [here.](
    https://scikit-optimize.github.io/stable/modules/generated/skopt.optimizer.Optimizer.html#skopt.optimizer.Optimizer)

    Example:
        ``` python
        my_pipe = Hyperpipe('skopt_example',
                            optimizer='sk_opt',
                            optimizer_params={'n_configurations': 25,
                                              'acq_func': 'LCB',
                                              'acq_func_kwargs': {'kappa': 1.96}},
                            ...)
        ```

    """
    def __init__(self,
                 n_configurations: int = 20,
                 n_initial_points: int = 10,
                 limit_in_minutes: Union[float, None] = None,
                 base_estimator: Union[str, sklearn.base.RegressorMixin] = "ET",
                 initial_point_generator: str = "random",
                 acq_func: str = 'gp_hedge',
                 acq_func_kwargs: dict = None):
        """
        Initialize the object.

        Parameters:
            n_configurations:
                Number of configurations to be calculated.

            n_initial_points:
                Number of evaluations with initialization points
                before approximating it with `base_estimator`.

            limit_in_minutes:
                Total time in minutes.

            base_estimator:
                Estimator for returning std(Y | x) along with E[Y | x].

            initial_point_generator:
                Generator for initial points.

            acq_func:
                Function to minimize over the posterior distribution.

            acq_func_kwargs:
                Additional arguments to be passed to the acquisition function.

        """
        self.metric_to_optimize = ''
        self.n_configurations = n_configurations
        self.n_initial_points = n_initial_points
        self.base_estimator = base_estimator
        self.initial_point_generator = initial_point_generator
        self.acq_func = acq_func
        self.acq_func_kwargs = acq_func_kwargs
        self.limit_in_minutes = limit_in_minutes
        self.start_time, self.end_time = None, None

        self.optimizer = None
        self.maximize_metric = None
        self.hyperparameter_list = []
        self.constant_dictionary = {}
        self.ask = self.ask_generator()

    def ask_generator(self) -> Generator:
        """
        Generator for new configs - ask method.

        Returns:
            Yields the next config.

        """
        if self.start_time is None and self.limit_in_minutes is not None:
            self.start_time = datetime.datetime.now()
            self.end_time = self.start_time + datetime.timedelta(minutes=self.limit_in_minutes)
        if self.optimizer is None:
            yield {}
        else:
            for i in range(self.n_configurations):
                next_config_list = self.optimizer.ask()
                next_config_dict = {self.hyperparameter_list[number]:
                                    self._convert_to_native(value) for number, value in enumerate(next_config_list)}
                if self.limit_in_minutes is None or datetime.datetime.now() < self.end_time:
                    yield next_config_dict
                else:
                    yield {}
                    break

    def prepare(self, pipeline_elements: list, maximize_metric: bool) -> None:
        """
        Initializes hyperparameter search with scikit-optimize.

        Assembles all hyperparameters of the list of PipelineElements
        in order to prepare the hyperparameter space.
        Hyperparameters can be accessed via pipe_element.hyperparameters.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create the hyperparameter space.

            maximize_metric:
                Boolean to distinguish between score and error.

        """
        self.start_time = None
        self.optimizer = None
        self.hyperparameter_list = []
        self.maximize_metric = maximize_metric

        # build skopt space
        space = []
        for pipe_element in pipeline_elements:
            if pipe_element.__class__.__name__ == 'Switch':
                error_msg = 'Scikit-Optimize cannot operate in the specified hyperparameter space with a Switch ' \
                            'element. We recommend the use of SMAC.'
                logger.error(error_msg)
                raise ValueError(error_msg)

            if hasattr(pipe_element, 'hyperparameters'):
                for name, value in pipe_element.hyperparameters.items():
                    # if we only have one value we do not need to optimize
                    if isinstance(value, list) and len(value) < 2:
                        self.constant_dictionary[name] = value[0]
                        continue
                    if isinstance(value, PhotonCategorical) and len(value.values) < 2:
                        self.constant_dictionary[name] = value.values[0]
                        continue
                    skopt_param = self._convert_photonai_to_skopt_space(value, name)
                    if skopt_param is not None:
                        space.append(skopt_param)

        if self.constant_dictionary:
            msg = "PHOTONAI has detected some one-valued params in your hyperparameters. Pleas use the kwargs for " \
                  "constant values. This run ignores following settings: " + str(self.constant_dictionary.keys())
            logger.warning(msg)
            warnings.warn(msg)

        if len(space) == 0:
            msg = "Did not find any hyperparameter to convert into skopt space."
            logger.warning(msg)
            warnings.warn(msg)
        else:
            self.optimizer = Optimizer(space,
                                       base_estimator=self.base_estimator,
                                       n_initial_points=self.n_initial_points,
                                       initial_point_generator=self.initial_point_generator,
                                       acq_func=self.acq_func,
                                       acq_func_kwargs=self.acq_func_kwargs)
        self.ask = self.ask_generator()

    def tell(self, config: dict, performance: float) -> None:
        """
        Provide a config result to calculate new ones.

        Parameters:
            config:
                The configuration that has been trained and tested.

            performance:
                Metrics about the configuration's generalization capabilities.

        """
        # convert dictionary to list in correct order
        if self.optimizer is not None:
            config_values = [config[name] for name in self.hyperparameter_list]
            best_config_metric_performance = performance
            if self.maximize_metric:
                best_config_metric_performance = -best_config_metric_performance
            self.optimizer.tell(config_values, best_config_metric_performance)

    def _convert_photonai_to_skopt_space(self, hyperparam: Union[PhotonHyperparam, list], name: str) -> Dimension:
        self.hyperparameter_list.append(name)
        if isinstance(hyperparam, PhotonCategorical) or isinstance(hyperparam, BooleanSwitch):
            return skoptCategorical(hyperparam.values, name=name)
        elif isinstance(hyperparam, list):
            return skoptCategorical(hyperparam, name=name)
        elif isinstance(hyperparam, FloatRange):
            if hyperparam.range_type == 'linspace':
                return Real(hyperparam.start, hyperparam.stop, name=name, prior='uniform')
            elif hyperparam.range_type == 'logspace':
                return Real(hyperparam.start, hyperparam.stop, name=name, prior='log-uniform')
            else:
                msg = "The hyperparam.range_type "+hyperparam.range_type+" is not supported by scikit-optimize."
                logger.error(msg)
                raise ValueError(msg)
        elif isinstance(hyperparam, IntegerRange):
            return Integer(hyperparam.start, hyperparam.stop, name=name)

        msg = "Cannot convert hyperparameter " + str(hyperparam) + ". " \
              "Supported types: Categorical, IntegerRange, FloatRange, list."
        logger.error(msg)
        raise ValueError(msg)

    @staticmethod
    def _convert_to_native(obj):
        # check if we have a numpy object, if so convert it to python native
        if type(obj).__module__ == np.__name__:
            return obj.item()
        else:
            return obj



#
#
# import warnings
# from os.path import join, dirname, realpath
# from pathlib import Path
# from photonai.helper.photon_base_test import PhotonBaseTest
# warnings.filterwarnings("ignore", category=DeprecationWarning)
# warnings.filterwarnings("ignore", category=RuntimeWarning)
# warnings.filterwarnings("ignore", category=FutureWarning)
#
#
# class TestRunExamples(PhotonBaseTest):
#
#     @classmethod
#     def setUpClass(cls) -> None:
#         cls.file = __file__
#         super(TestRunExamples, cls).setUpClass()
#
#     def setUp(self):
#         self.examples_folder = Path(dirname(realpath(__file__))).parent.parent.joinpath('examples')
#
#     def test_keras_dnn_regression(self):
#         exec(open(join(self.examples_folder, "neural_networks/keras_dnn_regression.py")).read(), locals(), globals())
#
#     def test_multi_layer_perceptron_classifier(self):
#         exec(open(join(self.examples_folder, "neural_networks/multi_layer_perceptron_classifier.py")).read(), locals(), globals())
#
#     def test_keras_dnn_callbacks(self):
#         exec(open(join(self.examples_folder, "neural_networks/keras_dnn_callbacks.py")).read(), locals(), globals())
#
#     def test_keras_cnn_classification(self):
#         exec(open(join(self.examples_folder, "neural_networks/keras_cnn_classification.py")).read(), locals(), globals())
#
#     def test_keras_dnn_multiclass_classification(self):
#         exec(open(join(self.examples_folder, "neural_networks/keras_dnn_multiclass_classification.py")).read(), locals(), globals())
#
#     def test_regression_with_constraints(self):
#         exec(open(join(self.examples_folder, "advanced/regression_with_constraints.py")).read(), locals(), globals())
#
#     def test_multiclass_classification(self):
#         exec(open(join(self.examples_folder, "advanced/multiclass_classification.py")).read(), locals(), globals())
#
#     def test_sample_pairing_example(self):
#         exec(open(join(self.examples_folder, "advanced/sample_pairing_example.py")).read(), locals(), globals())
#
#     def test_imbalanced_data(self):
#         exec(open(join(self.examples_folder, "advanced/imbalanced_data.py")).read(), locals(), globals())
#
#     def test_confounder_removal_example(self):
#         exec(open(join(self.examples_folder, "advanced/confounder_removal_example.py")).read(), locals(), globals())
#
#     def test_group_split(self):
#         exec(open(join(self.examples_folder, "advanced/group_split.py")).read(), locals(), globals())
#
#     def test_svc_ensemble(self):
#         exec(open(join(self.examples_folder, "advanced/svc_ensemble.py")).read(), locals(), globals())
#
#     def test_estimator_stack_voting(self):
#         exec(open(join(self.examples_folder, "advanced/estimator_stack_voting.py")).read(), locals(), globals())
#
#     def test_callbacks(self):
#         exec(open(join(self.examples_folder, "advanced/callbacks.py")).read(), locals(), globals())
#
#     def test_additional_data(self):
#         exec(open(join(self.examples_folder, "advanced/additional_data.py")).read(), locals(), globals())
#
#     def test_permutation_test(self):
#         exec(open(join(self.examples_folder, "advanced/permutation_test.py")).read(), locals(), globals())
#
#     def test_register_elements(self):
#         exec(open(join(self.examples_folder, "advanced/register_elements.py")).read(), locals(), globals())
#
#     def test_feature_selection(self):
#         exec(open(join(self.examples_folder, "advanced/feature_selection.py")).read(), locals(), globals())
#
#     def test_custom_estimator(self):
#         exec(open(join(self.examples_folder, "advanced/custom_elements/custom_estimator.py")).read(), locals(), globals())
#
#     def test_custom_transformer(self):
#         exec(open(join(self.examples_folder, "advanced/custom_elements/custom_transformer.py")).read(), locals(), globals())
#
#     def test_regression(self):
#         exec(open(join(self.examples_folder, "getting_started/regression.py")).read(), locals(), globals())
#
#     def test_no_outer_cv_default_pipe(self):
#         exec(open(join(self.examples_folder, "getting_started/no_outer_cv_default_pipe.py")).read(), locals(), globals())
#
#     def test_load_hyperpipe_from_json(self):
#         exec(open(join(self.examples_folder, "getting_started/load_hyperpipe_from_json.py")).read(), locals(), globals())
#
#     def test_batching_elements(self):
#         exec(open(join(self.examples_folder, "getting_started/batching_elements.py")).read(), locals(), globals())
#
#     def test_pipeline_branches(self):
#         exec(open(join(self.examples_folder, "getting_started/pipeline_branches.py")).read(), locals(), globals())
#
#     def test_stack(self):
#         exec(open(join(self.examples_folder, "getting_started/stack.py")).read(), locals(), globals())
#
#     def test_classification(self):
#         exec(open(join(self.examples_folder, "getting_started/classification.py")).read(), locals(), globals())
#
#     def test_data_integration(self):
#         exec(open(join(self.examples_folder, "getting_started/data_integration.py")).read(), locals(), globals())
#
#     def test_jmlr_example(self):
#         exec(open(join(self.examples_folder, "getting_started/jmlr_example.py")).read(), locals(), globals())
#
#     def test_switch(self):
#         exec(open(join(self.examples_folder, "getting_started/switch.py")).read(), locals(), globals())
#
#     def test_classifier_ensemble(self):
#         exec(open(join(self.examples_folder, "getting_started/classifier_ensemble.py")).read(), locals(), globals())
#
#     def test_skopt_example(self):
#         exec(open(join(self.examples_folder, "optimizer/skopt_example.py")).read(), locals(), globals())
#
#     def test_comparisson_example(self):
#         exec(open(join(self.examples_folder, "optimizer/comparisson_example.py")).read(), locals(), globals())
#
#     def test_smac_example(self):
#         exec(open(join(self.examples_folder, "optimizer/smac_example.py")).read(), locals(), globals())
#
#     def test_results_example(self):
#         exec(open(join(self.examples_folder, "results/results_example.py")).read(), locals(), globals())
#
#     def test_mongodb(self):
#         exec(open(join(self.examples_folder, "results/mongodb.py")).read(), locals(), globals())



import numpy as np
import statsmodels.api as sm
from scipy.linalg import cholesky
from scipy.stats import norm
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold, ShuffleSplit
from sklearn.preprocessing import StandardScaler

from photonai.base import Hyperpipe, PipelineElement
from photonai.helper.photon_base_test import PhotonBaseTest


class ConfounderRemovalTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(ConfounderRemovalTests, cls).setUpClass()

    def setUp(self):

        super(ConfounderRemovalTests, self).setUp()
        self.X, self.y = load_breast_cancer(return_X_y=True)
        self.X_train = self.X[:100]
        self.y_train = self.y[:100]
        self.shuffle_split = ShuffleSplit(test_size=0.2, n_splits=1, random_state=15)
        self.pipe = Hyperpipe("confounder_pipe",
                              outer_cv=self.shuffle_split, inner_cv= KFold(n_splits=3, shuffle=True, random_state=15),
                              metrics=["accuracy"], best_config_metric="accuracy", project_folder=self.tmp_folder_path)
        self.pipe += PipelineElement("StandardScaler")
        self.cr = PipelineElement("ConfounderRemoval")
        self.pipe += self.cr
        self.pipe += PipelineElement("SVC")
        self.random_confounders = np.random.randn(self.X.shape[0], 1)

        # do confounder removal by hand
        self.multiple_confounders = np.random.randn(self.X.shape[0], 2) * 10
        ols_confounder = sm.add_constant(self.multiple_confounders)
        self.X_transformed = np.empty(self.X.shape)
        for i in range(self.X.shape[1]):
            # fit
            model = sm.OLS(endog=np.squeeze(self.X[:, i]), exog=ols_confounder).fit()
            # transform
            self.X_transformed[:, i] = np.asarray(np.squeeze(self.X[:, i]) - np.matmul(ols_confounder, np.squeeze(model.params)))

        # prepare caching
        self.X_train_transformed = np.empty(self.X_train.shape)
        self.confounder_train = self.multiple_confounders[:100]
        ols_confounder_train = sm.add_constant(self.confounder_train)
        for i in range(self.X_train.shape[1]):
            # fit
            model = sm.OLS(endog=np.squeeze(self.X_train[:, i]), exog=ols_confounder_train).fit()
            # transform
            self.X_train_transformed[:, i] = np.asarray(
                np.squeeze(self.X_train[:, i]) - np.matmul(ols_confounder_train, np.squeeze(model.params)))

        # prepare confounder removal with standardization of covariates
        scaled_covs = list()
        # standardize covariates
        for cov in self.multiple_confounders.T:
            scaler = StandardScaler()
            scaled_covs.append(scaler.fit_transform(cov.reshape(-1, 1)).squeeze())
        scaled_covs = np.asarray(scaled_covs).T
        scaled_covs = sm.add_constant(scaled_covs)
        self.X_transformed_standardized = np.empty(self.X.shape)
        for i in range(self.X.shape[1]):
            # fit
            model = sm.OLS(endog=np.squeeze(self.X[:, i]), exog=scaled_covs).fit()
            # transform
            self.X_transformed_standardized[:, i] = np.asarray(
                np.squeeze(self.X[:, i]) - np.matmul(scaled_covs, np.squeeze(model.params)))

        # prepare statistical testing of confounder removal
        # Generate samples from three independent normally distributed random
        # variables (with mean 0 and std. dev. 1).
        x = norm.rvs(size=(4, 300))

        # desired covariance matrix
        r = np.array([
            [1, .9, .9, .9],
            [.9, 1, .9, .9],
            [.9, .9, 1, .9],
            [.9, .9, .9, 1],
        ])
        c = cholesky(r, lower=True)

        # convert the data to correlated random variables
        self.z = np.dot(c, x).T

    def test_confounder_removal_statistically(self):
        cr = PipelineElement("ConfounderRemoval", {}, standardize_covariates=False)
        cr.fit(self.z[:, 1:3], self.z[:, 0], **{'confounder': self.z[:, 3]})

        # use transform to write data to cache
        z_transformed = cr.transform(self.z[:, 1:3], **{'confounder': self.z[:,3]})
        corr = np.corrcoef(np.concatenate([self.z[:, 0].reshape(-1, 1), z_transformed[0],
                                           self.z[:, 3].reshape(-1, 1)], axis=1), rowvar=False)
        # correlation between target and feature should be lower than 0.25 in this case
        # correlation between covariate and feature should be near zero
        self.assertLess(corr[1, 0], 0.3)
        self.assertLess(corr[2, 0], 0.3)
        self.assertAlmostEqual(corr[3, 1], 0)
        self.assertAlmostEqual(corr[3, 2], 0)

    def test_multiple_confounders(self):
        self.cr.fit(self.X, self.y, **{'confounder': self.multiple_confounders})
        X_transformed = self.cr.transform(self.X, **{'confounder': self.multiple_confounders})
        np.testing.assert_array_almost_equal(X_transformed[0], self.X_transformed)

    def test_standardize_covariates(self):
        self.cr.fit(self.X, self.y, **{'confounder': self.multiple_confounders})
        X_transformed = self.cr.transform(self.X, **{'confounder': self.multiple_confounders})
        np.testing.assert_array_almost_equal(X_transformed[0], self.X_transformed_standardized)

    def test_use(self):
        self.pipe.fit(self.X, self.y, **{'confounder': self.random_confounders})
        trans_data = self.pipe.transform(self.X, **{'confounder': self.random_confounders})

    def test_dimensions(self):
        with self.assertRaises(ValueError):
            self.cr.fit(self.X, self.y, confounder=np.random.randn(self.X.shape[0]-10, 2))

    def test_key_error(self):
        with self.assertRaises(KeyError):
            self.cr.fit(self.X, self.y, covariate=np.random.randn(self.X.shape[0]-10, 2))



# content by J. Brownlee:
# https://machinelearningmastery.com/cnn-models-for-human-activity-recognition-time-series-classification/
# HAR-Dataset: https://archive.ics.uci.edu/ml/datasets/human+activity+recognition+using+smartphones

from numpy import dstack
from pandas import read_csv
import numpy as np


# load a single file as a numpy array
def load_file(filepath):
    dataframe = read_csv(filepath, header=None, delim_whitespace=True)
    return dataframe.values


# load a list of files and return as a 3d numpy array
def load_group(filenames, prefix=''):
    loaded = list()
    for name in filenames:
        data = load_file(prefix + name)
        loaded.append(data)
    # stack group so that features are the 3rd dimension
    loaded = dstack(loaded)
    return loaded


# load a dataset group, such as train or test
def load_dataset_group(group, prefix=''):
    filepath = prefix + group + '/Inertial Signals/'
    # load all 9 files as a single array
    filenames = list()
    # total acceleration
    filenames += ['total_acc_x_' + group + '.txt', 'total_acc_y_' + group + '.txt', 'total_acc_z_' + group + '.txt']
    # body acceleration
    filenames += ['body_acc_x_' + group + '.txt', 'body_acc_y_' + group + '.txt', 'body_acc_z_' + group + '.txt']
    # body gyroscope
    filenames += ['body_gyro_x_' + group + '.txt', 'body_gyro_y_' + group + '.txt', 'body_gyro_z_' + group + '.txt']
    # load input data
    X = load_group(filenames, filepath)
    # load class output
    y = load_file(prefix + group + '/y_' + group + '.txt')
    return X, y


# load the dataset, returns train and test X and y elements
def load_har(prefix=''):
    # load all train
    trainX, trainy = load_dataset_group('train', prefix + '/')
    print(trainX.shape, trainy.shape)
    # load all test
    testX, testy = load_dataset_group('test', prefix + '/')
    print(testX.shape, testy.shape)
    # zero-offset class values
    trainy = trainy - 1
    testy = testy - 1
    print(trainX.shape, trainy.shape, testX.shape, testy.shape)
    return np.concatenate((trainX, testX), axis=0), np.concatenate((trainy, testy), axis=0)



import numpy as np
from sklearn.linear_model import Lasso
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_selection import f_regression, f_classif, SelectPercentile, VarianceThreshold

from photonai.photonlogger.logger import logger


class FRegressionFilterPValue(BaseEstimator, TransformerMixin):
    """Feature Selection for Regression - p-value based.

    Fit f_regression and select all columns
    when p_value of column < p_threshold.

    """
    _estimator_type = "transformer"

    def __init__(self, p_threshold: float = .05):
        """
        Initialize the object.

        Parameters:
            p_threshold:
                Upper bound for p_values.

        """
        self.p_threshold = p_threshold
        self.selected_indices = []
        self.n_original_features = None

    def fit(self, X: np.ndarray, y: np.ndarray):
        """Calculation of the important columns.

        Apply f_regression on input X, y to generate p_values.
        selected_indices = all p_value(columns) < p_threshold.

        Parameters:
            X:
                The input samples of shape [n_samples, n_original_features]

            y:
                The input targets of shape [n_samples, 1]

        """
        self.n_original_features = X.shape[1]
        _, p_values = f_regression(X, y)
        self.selected_indices = np.where(p_values < self.p_threshold)[0]
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Reduced input X to selected_columns.

        Parameters:
            X
                The input samples of shape [n_samples, n_original_features]

        Returns:
            Column-filtered array of shape [n_samples, n_selected_features].

        """
        return X[:, self.selected_indices]

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """Reverse to original dimension.

        Parameters:
            X:
                The input samples of shape [n_samples, n_selected_features].

        Raises:
            ValueError: If input X has a different shape than during fitting.

        Returns:
            Array of shape [n_samples, n_original_features]
            with columns of zeros inserted where features would have
            been removed.

        """
        if X.shape[1] != len(self.selected_indices):
            msg = "X has a different shape than during fitting."
            logger.error(msg)
            raise ValueError(msg)

        Xt = np.zeros((X.shape[0], self.n_original_features))
        Xt[:, self.selected_indices] = X
        return Xt


class FRegressionSelectPercentile(BaseEstimator, TransformerMixin):
    """Feature Selection for regression data - percentile based.

    Apply VarianceThreshold -> SelectPercentile to data.
    SelectPercentile based on f_regression and parameter percentile.

    """
    _estimator_type = "transformer"

    def __init__(self, percentile: float = 10):
        """
        Initialize the object.

        Parameters:
            percentile:
                Percent of features to keep.

        """
        self.var_thres = VarianceThreshold()
        self.percentile = percentile
        self.my_fs = None

    def fit(self, X, y):
        X = self.var_thres.fit_transform(X)
        self.my_fs = SelectPercentile(score_func=f_regression, percentile=self.percentile)
        self.my_fs.fit(X,y)
        return self

    def transform(self, X):
        X = self.var_thres.transform(X)
        return self.my_fs.transform(X)

    def inverse_transform(self, X):
        Xt = self.my_fs.inverse_transform(X)
        return self.var_thres.inverse_transform(Xt)


class FClassifSelectPercentile(BaseEstimator, TransformerMixin):
    """Feature Selection for classification data - percentile based.

    Apply VarianceThreshold -> SelectPercentile to data.
    SelectPercentile based on f_classif and parameter percentile.

    """
    _estimator_type = "transformer"

    def __init__(self, percentile: float = 10):
        """
        Initialize the object.

        Parameters:
            percentile:
                Percent of features to keep.

        """
        self.var_thres = VarianceThreshold()
        self.percentile = percentile
        self.my_fs = None

    def fit(self, X, y):
        X = self.var_thres.fit_transform(X)
        self.my_fs = SelectPercentile(score_func=f_classif, percentile=self.percentile)
        self.my_fs.fit(X, y)
        return self

    def transform(self, X):
        X = self.var_thres.transform(X)
        return self.my_fs.transform(X)

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """Reverse to original dimension.

        1. SelectPercentile.inverse_transform
        2. VarianceThreshold.inverse_transform

        Parameters:
            X:
                The input samples of shape [n_samples, n_selected_features].

        Returns:
            Array of shape [n_samples, n_original_features]
            with columns of zeros inserted where features would have
            been removed.

        """
        Xt = self.my_fs.inverse_transform(X)
        return self.var_thres.inverse_transform(Xt)


class ModelSelector(BaseEstimator, TransformerMixin):
    """Model Selector - based on feature_importance.

    Apply feature selection on specific estimator
    and its importance scores.

     """
    _estimator_type = "transformer"

    def __init__(self, estimator_obj: BaseEstimator, threshold: float = 1e-5, percentile: bool = False):
        """
        Initialize the object.

        Parameters:
            estimator_obj:
                Estimator with fit/tranform and possibility of feature_importance.

            threshold:
                If percentile == True:
                    Lower Bound for required importance score to keep.
                If percentile == True:
                    percentage to keep (ordered features by feature_importance)

            percentile:
                Percent of features to keep.

        """
        self.threshold = threshold
        self.estimator_obj = estimator_obj
        self.selected_indices = []
        self.percentile = percentile
        self.importance_scores = []
        self.n_original_features = None

    def _get_feature_importances(self, estimator, norm_order=1):
        """Retrieve or aggregate feature importances from estimator"""
        importances = getattr(estimator, "feature_importances_", None)

        if importances is None and hasattr(estimator, "coef_"):
            if estimator.coef_.ndim == 1:
                importances = np.abs(estimator.coef_)

            else:
                importances = np.linalg.norm(estimator.coef_, axis=0,
                                             ord=norm_order)

        elif importances is None:
            raise ValueError(
                "The underlying estimator %s has no `coef_` or "
                "`feature_importances_` attribute. Either pass a fitted estimator"
                " to SelectFromModel or call fit before calling transform."
                % estimator.__class__.__name__)

        return importances

    def fit(self, X, y=None, **kwargs):
        self.n_original_features = X.shape[1]
        # 1. fit estimator
        self.estimator_obj.fit(X, y)
        # penalty = "l1"
        self.importance_scores = self._get_feature_importances(self.estimator_obj)

        if not self.percentile:
            self.selected_indices = np.where(self.importance_scores >= self.threshold)[0]
        else:
            # Todo: works only for binary classification, not for multiclass
            if self.threshold > 1:
                raise ValueError("Threshold should not be greater than 1")
            ordered_importances = np.sort(self.importance_scores)
            if isinstance(X, list):
                X = np.array(X)
            index = int(np.floor((1-self.threshold) * X.shape[1]))
            percentile_thres = ordered_importances[index]
            self.selected_indices = np.where(self.importance_scores >= percentile_thres)[0]
            # Todo: sortieren und Grenze definieren und dann np.where
            pass
        return self

    def transform(self, X, y=None, **kwargs):

        if isinstance(X, list):
            X = np.array(X)

        X_new = X[:, self.selected_indices]

        # if no features were selected raise error
        if X_new.shape[1] == 0:
            print("No Features were selected from model, using all features")
            return X
        return X_new

    def inverse_transform(self, X):
        if X.shape[1] != len(self.selected_indices):
            msg = "X has a different shape than during fitting."
            logger.error(msg)
            raise ValueError(msg)
        Xt = np.zeros((X.shape[0], self.n_original_features))
        Xt[:, self.selected_indices] = X
        return Xt

    def set_params(self, **params):
        if 'threshold' in params:
            self.threshold = params['threshold']
            params.pop('threshold')
        self.estimator_obj.set_params(**params)

    def get_params(self, deep=True):
        return self.estimator_obj.get_params(deep)


class LassoFeatureSelection(BaseEstimator, TransformerMixin):
    """Lasso based feature selection - based on feature_importance.

    Apply Lasso to ModelSelection.

    """
    def __init__(self, percentile: float = 0.3, alpha: float = 1., **kwargs):
        """
        Initialize the object.

        Parameters:
            percentile: bool, default=False
                Percent of features to keep.

            alpha: float, default=1.
                Weighting parameter for Lasso.

            **kwargs:
                Passed to Lasso object.

        """
        self.percentile = percentile
        self.alpha = alpha
        self.model_selector = None
        self.Lasso_kwargs = kwargs
        self.needs_covariates=False
        self.needs_y = False

    def fit(self, X, y=None, **kwargs):
        self.model_selector = ModelSelector(Lasso(alpha=self.alpha, **self.Lasso_kwargs),
                                            threshold=self.percentile, percentile=True)

        self.model_selector.fit(X, y, **kwargs)
        return self

    def transform(self, X, y=None, **kwargs):
        selected_features = self.model_selector.transform(X, y, **kwargs)
        return selected_features

    def set_params(self, **params):
        super(LassoFeatureSelection, self).set_params(**params)

    def inverse_transform(self, X, y=None, **kwargs):
        return self.model_selector.inverse_transform(X)



import unittest
import numpy as np
from sklearn.datasets import load_breast_cancer
from photonai.modelwrapper.PhotonOneClassSVM import PhotonOneClassSVM
from photonai.modelwrapper.RangeRestrictor import RangeRestrictor
from photonai.modelwrapper.PhotonMLPClassifier import PhotonMLPClassifier


class ModelWrapperTests(unittest.TestCase):

    def setUp(self):
        self.X, self.y = load_breast_cancer(True)

    def test_photon_mlp(self):
        mlp = PhotonMLPClassifier()
        mlp.fit(self.X, self.y)
        mlp.predict(self.X)
        mlp.predict_proba(self.X)

    def test_photon_one_class_svm(self):
        osvm = PhotonOneClassSVM()
        osvm.fit(self.X, self.y)
        osvm.predict(self.X)
        osvm.score(self.X, self.y)
        osvm.get_params()
        osvm.set_params(**{'kernel': 'linear'})

    def test_range_restrictor(self):
        rr = RangeRestrictor()
        rr.fit(self.X, self.y)
        pred = rr.predict(self.X)
        self.assertEqual(np.min(pred), 0)
        self.assertEqual(np.max(pred), 100)



import datetime
import random
from typing import Union, Generator

from photonai.optimization.base_optimizer import PhotonSlaveOptimizer
from photonai.photonlogger.logger import logger


class RandomSearchOptimizer(PhotonSlaveOptimizer):
    """Random search optimizer.

    Searches for the best configuration by randomly
    testing hyperparameter combinations without any grid.

    """
    def __init__(self, limit_in_minutes: Union[float, None] = 60, n_configurations: Union[int, None] = None):
        """
        Initialize the object.
        One of limit_in_minutes or n_configurations must differ from None.

        Parameters:
            limit_in_minutes:
                Total time in minutes.

            n_configurations:
                Number of configurations to be calculated.

        """
        self.pipeline_elements = None
        self.parameter_iterable = None
        self.ask = self.next_config_generator()
        self.n_configurations = None

        if not limit_in_minutes or limit_in_minutes <= 0:
            self.limit_in_minutes = None
        else:
            self.limit_in_minutes = limit_in_minutes
        self.start_time = None
        self.end_time = None

        if not n_configurations or n_configurations <= 0:
            self.n_configurations = None
        else:
            self.n_configurations = n_configurations
        self.k_configutration = 0  # use k++ until k==n: break

        if self.n_configurations is None and self.limit_in_minutes is None:
            msg = "No stopping criteria for RandomSearchOptimizer."
            logger.error(msg)
            raise ValueError(msg)

    def prepare(self, pipeline_elements: list, maximize_metric: bool) -> None:
        """
        Initializes grid free random hyperparameter search.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create the hyperparameter space.

            maximize_metric:
                Boolean to distinguish between score and error.

        """
        self.start_time = None
        self.pipeline_elements = pipeline_elements
        self.ask = self.next_config_generator()

    def next_config_generator(self) -> Generator:
        """
        Generator for new configs - ask method.

        Returns:
            Yields the next config.

        """
        while True:
            _ = (yield self._generate_config())
            self.k_configutration += 1
            if self.limit_in_minutes:
                if self.start_time is None:
                    self.start_time = datetime.datetime.now()
                    self.end_time = self.start_time + datetime.timedelta(minutes=self.limit_in_minutes)

                if datetime.datetime.now() >= self.end_time:
                    return

            if self.n_configurations:
                if self.k_configutration >= self.n_configurations:
                    return

    def _generate_config(self):
        config = {}
        for p_element in self.pipeline_elements:
            for h_key, h_value in p_element.hyperparameters.items():
                if isinstance(h_value, list):
                    config[h_key] = random.choice(h_value)
                else:
                    config[h_key] = h_value.get_random_value()
        return config
